#!/usr/bin/env python3

import unittest
import os, sys
import math

sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.abspath(__file__))))

from shapely.geometry import LineString, Point, Polygon

import hsm_nibble.geometry as geometry


ACCURACY = 0.05


class BaseTests(unittest.TestCase):
    def verify_arc(self, arc):
        self.verify_arc_points(arc)
        self.verify_arc_angles(arc)

    def verify_arc_points(self, arc):
        self.assertIsNotNone(arc.radius)
        self.assertIsNotNone(arc.start)
        self.assertIsNotNone(arc.end)
        self.assertIsNotNone(arc.path)
        self.assertIsNotNone(arc.origin)

        self.assertEqual(arc.start, Point(arc.path.coords[0]))
        self.assertEqual(arc.end, Point(arc.path.coords[-1]))

        for point in arc.path.coords:
            self.assertAlmostEqual(arc.radius, arc.origin.distance(Point(point)))

    def verify_arc_angles(self, arc):
        self.assertIsNotNone(arc.start_angle)
        self.assertIsNotNone(arc.span_angle)
        self.assertIsNotNone(arc.winding_dir)

        start_angle_observed = math.atan2(arc.start.x - arc.origin.x, arc.start.y - arc.origin.y)
        start_angle_observed = start_angle_observed % (math.pi * 2)
        end_angle_observed = math.atan2(arc.end.x - arc.origin.x, arc.end.y - arc.origin.y)
        end_angle_observed = end_angle_observed % (math.pi * 2)

        self.assertAlmostEqual(start_angle_observed, arc.start_angle)
        self.assertAlmostEqual(end_angle_observed, (arc.start_angle + arc.span_angle) % (2 * math.pi))

        self.assertAlmostEqual(arc.path.length, arc.radius * abs(arc.span_angle), places=1)

        if arc.winding_dir == geometry.ArcDir.CW:
            self.assertGreater(arc.span_angle, 0)
        elif arc.winding_dir == geometry.ArcDir.CCW:
            self.assertLess(arc.span_angle, 0)


class TestCircle(unittest.TestCase):
    def test_create(self):
        """ Check circles are create in a sane way. """
        origin = Point(10, 10)
        radius = 7

        circle = geometry.create_circle(origin = origin, radius = radius)

        self.assertEqual(circle.origin, origin)
        self.assertEqual(circle.radius, radius)
        self.assertIsNone(circle.start)
        self.assertIsNone(circle.end)
        self.assertEqual(circle.start_angle, 0)
        self.assertEqual(circle.span_angle, 2 * math.pi)
        self.assertIsNotNone(circle.path)
        self.assertEqual(circle.path.coords[0], circle.path.coords[-1])
        self.assertNotEqual(circle.path.coords[0], circle.path.coords[1])

        mid_point = circle.path.coords[int(len(circle.path.coords)/2)]
        length_to_mid = LineString([circle.path.coords[0], mid_point]).length
        self.assertEqual(length_to_mid, 2 * radius)

class TestCreateArc(BaseTests):
    def test_create_cw(self):
        """ Check create_arc() method works in a sane way for ClockWise arcs. """
        origin = Point(100, 100)
        radius = 5
        start_angle = geometry.math.pi / 2
        span_angle = geometry.math.pi / 2
        winding_dir = geometry.ArcDir.CW

        arc = geometry.create_arc(origin, radius, start_angle, span_angle, winding_dir)

        self.verify_arc(arc)

    def test_create_ccw(self):
        """ Check create_arc() method works in a sane way for CounterClockWise arcs. """
        origin = Point(-123, -123)
        radius = 5
        start_angle = 3 * geometry.math.pi / 2
        span_angle = -geometry.math.pi / 4
        winding_dir = geometry.ArcDir.CCW

        arc = geometry.create_arc(origin, radius, start_angle, span_angle, winding_dir)

        self.verify_arc(arc)

    def test_create_cw_full_circle(self):
        """ Check create_arc() method works in a sane way for CounterClockWise arcs. """
        origin = Point(123, 123)
        radius = 5
        start_angle = 0
        span_angle = math.pi * 2
        winding_dir = geometry.ArcDir.CW

        arc = geometry.create_arc(origin, radius, start_angle, span_angle, winding_dir)

        self.verify_arc(arc)

    def test_create_ccw_full_circle(self):
        """ Check create_arc() method works in a sane way for CounterClockWise arcs. """
        origin = Point(-100, -100)
        radius = 5
        start_angle = 0
        span_angle = -math.pi * 2
        winding_dir = geometry.ArcDir.CCW

        arc = geometry.create_arc(origin, radius, start_angle, span_angle, winding_dir)

        self.verify_arc(arc)


class TestArc(BaseTests):
    def test_create_arc_from_path_CW(self):
        """ Create a Clockwise arc. """
        origin = Point(123.45, 123.45)
        radius = 7
        winding_dir = geometry.ArcDir.CW

        # This path will start at the coordinates: (radius, 0)
        path = list(geometry.create_circle(origin, radius).path.coords)

        path = path[int(len(path) / 8) : int(7 * len(path) / 8)]

        arc = geometry.create_arc_from_path(origin, LineString(path), radius)
        self.verify_arc_points(arc)
        arc = geometry.complete_arc(arc, winding_dir)

        self.assertEqual(arc.origin, origin)
        self.assertEqual(arc.radius, radius)
        self.assertEqual(arc.winding_dir, winding_dir)

        # We start this arc at the coordinate point: (radius, 0) with the origin at (0, 0).
        # Angle rotated from full circle's start/end point:
        expected_start_angle = (1 / 8) * (2 * math.pi)
        # Add to full circle's start angle.
        expected_start_angle = (math.pi / 2) + expected_start_angle

        self.assertLess(abs(arc.start_angle - expected_start_angle), ACCURACY * math.pi * 2 )

        self.verify_arc(arc)

    def test_create_arc_from_path_CCW(self):
        """ Create a Counter-Clockwise arc. """
        origin = Point(-123.45, -123.45)
        radius = 7
        winding_dir = geometry.ArcDir.CCW

        # This path will start at the coordinates: (radius, 0)
        path = list(geometry.create_circle(origin, radius).path.coords)

        path = path[int(len(path) / 8) : int(7 * len(path) / 8)]

        arc = geometry.create_arc_from_path(origin, LineString(path), radius)
        self.verify_arc_points(arc)
        arc = geometry.complete_arc(arc, winding_dir)

        self.assertEqual(arc.origin, origin)
        self.assertEqual(arc.radius, radius)
        self.assertEqual(arc.winding_dir, winding_dir)

        # We start this arc at the coordinate point: (radius, 0) with the origin at (0, 0).
        # Angle rotated from full circle's start/end point:
        expected_start_angle = (1 / 8) * (2 * math.pi)
        # Add to full circle's start angle.
        expected_start_angle = (math.pi / 2) - expected_start_angle
        self.assertLess(abs(arc.start_angle - expected_start_angle), ACCURACY * math.pi * 2 )

        self.verify_arc(arc)


class TestMirrorArc(BaseTests):
    def test_mirror_arc_from_path_CW(self):
        """ Mirror a Clockwise arc. """
        origin = Point(100, 100)
        radius = 5
        span_angle = geometry.math.pi / 2
        winding_dir = geometry.ArcDir.CW
        mirror_axis = 123

        for start_angle_multiplier in range(8):
            start_angle = start_angle_multiplier * math.pi * 2 / 8
            arc = geometry.create_arc(origin, radius, start_angle, span_angle, winding_dir)
            self.verify_arc(arc)

            mirrored_arc = geometry.mirror_arc(mirror_axis, arc)
            self.verify_arc(mirrored_arc)
            self.assertEqual(mirrored_arc.winding_dir, geometry.ArcDir.CCW)

    def test_mirror_arc_from_path_CCW(self):
        """ Mirror a Counter Clockwise arc. """
        origin = Point(100, 100)
        radius = 5
        span_angle = -geometry.math.pi / 2
        winding_dir = geometry.ArcDir.CCW
        mirror_axis = 123

        for start_angle_multiplier in range(8):
            start_angle = start_angle_multiplier * math.pi * 2 / 8
            arc = geometry.create_arc(origin, radius, start_angle, span_angle, winding_dir)
            self.verify_arc(arc)

            mirrored_arc = geometry.mirror_arc(mirror_axis, arc)
            self.verify_arc(mirrored_arc)
            self.assertEqual(mirrored_arc.winding_dir, geometry.ArcDir.CW)

    def test_mirror_arc_from_path_CW_force_dir_no_change(self):
        """ Try to mirror a Clockwise arc but no change in direction specified."""
        origin = Point(100, 100)
        radius = 5
        span_angle = geometry.math.pi / 2
        winding_dir = geometry.ArcDir.CW
        mirror_axis = 123
        start_angle = math.pi * 2 / 8

        arc = geometry.create_arc(origin, radius, start_angle, span_angle, winding_dir)
        self.verify_arc(arc)

        mirrored_arc = geometry.mirror_arc(mirror_axis, arc, geometry.ArcDir.CW)

        self.assertIs(arc, mirrored_arc)

    def test_mirror_arc_from_path_CW_force_dir_change(self):
        """ Mirror a Clockwise arc. Force winding_dir to a set state."""
        origin = Point(100, 100)
        radius = 5
        span_angle = geometry.math.pi / 2
        winding_dir = geometry.ArcDir.CW
        mirror_axis = 123

        for start_angle_multiplier in range(8):
            start_angle = start_angle_multiplier * math.pi * 2 / 8
            arc = geometry.create_arc(origin, radius, start_angle, span_angle, winding_dir)
            self.verify_arc(arc)

            mirrored_arc = geometry.mirror_arc(mirror_axis, arc, geometry.ArcDir.CCW)
            self.verify_arc(mirrored_arc)

if __name__ == '__main__':
    unittest.main()




#!/usr/bin/env python3
"""
Experimenting with CNC machining toolpaths.
This program is a demo which uses the main library file on test .dxf CAD files.

The demo shows how to calculate a HSM path on the outside of the shape to be cut.
"geometry.Pocket(...)" takes an "already_cut" parameter which represents the outer boundary of the
area to be cut. eg: The edge of the stock size.
"""

from typing import List, Tuple

import os
import sys
import math
import ezdxf
import matplotlib.pyplot as plt    # type: ignore
from shapely.affinity import rotate  # type: ignore
from shapely.geometry import MultiPolygon, Point, Polygon, LineString  # type: ignore

# This line is required if you want to use the local version of the code.
# If you have installed HSM_nibble via PIP it is not required.
sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.abspath(__file__))))

from hsm_nibble import dxf
from hsm_nibble import geometry

def display_start_point(toolpath, colour="purple"):
    plt.plot(toolpath.start_point.x, toolpath.start_point.y, marker='o', c=colour, markersize=10)

def display_outline(shapes, colour="blue"):
    """ Display the outline of the shape to be cut. """
    if shapes.type != "MultiPolygon":
        shapes = MultiPolygon([shapes])
    for shape in shapes.geoms:
        x, y = shape.exterior.xy
        plt.plot(x, y, c=colour, linewidth=2)

        for interior in shape.interiors:
            x, y = interior.xy
            plt.plot(x, y, c=colour, linewidth=2)

def display_voronoi(toolpath, colour="red"):
    """ Display the voronoi edges. These are equidistant from the shape's edges. """
    for edge in toolpath.voronoi.edges.values():
        x = []
        y = []
        for point in edge.coords:
            x.append(point[0])
            y.append(point[1])
        plt.plot(x, y, c="red", linewidth=4)
        plt.plot(x[0], y[0], 'x', c=colour)
        plt.plot(x[-1], y[-1], 'x', c=colour)

def display_visited_voronoi_edges(toolpath, colour="black"):
    """
    Display the voronoi edges that were used to calculate cut geometry.
    This should match the output of display_voronoi(...).
    """
    for edge in toolpath.visited_edges:
        edge_coords = toolpath.voronoi.edges[edge].coords
        x = []
        y = []
        for point in edge_coords:
            x.append(point[0])
            y.append(point[1])
        plt.plot(x, y, c=colour, linewidth=1)
        plt.plot(x[0], y[0], 'x', c=colour)
        plt.plot(x[-1], y[-1], 'x', c=colour)

def display_toolpath(
        toolpath, cut_colour="green", rapid_inside_colour="blue", rapid_outside_colour="orange"):
    # Display path.
    for element in toolpath.path:
        if type(element).__name__ == "Arc":
            x, y = element.path.xy
            if element.debug:
                plt.plot(x, y, c=element.debug, linewidth=3)
            else:
                plt.plot(x, y, c=cut_colour, linewidth=1)

        elif type(element).__name__ == "Line":
            x, y = element.path.xy
            if element.move_style == geometry.MoveStyle.RAPID_INSIDE:
                plt.plot(x, y, linestyle='--', c=rapid_inside_colour, linewidth=1)
            elif element.move_style == geometry.MoveStyle.RAPID_OUTSIDE:
                plt.plot(x, y, c=rapid_outside_colour, linewidth=1)
            else:
                assert element.move_style == geometry.MoveStyle.CUT
                plt.plot(x, y, linestyle='--', c=cut_colour, linewidth=1)

def display_entry_point(toolpath):
    if toolpath.starting_radius is None:
        # Entry circle not set.
        return

    starting_circle = geometry.create_circle(
            toolpath.start_point, toolpath.starting_radius).path
    x, y = starting_circle.xy

    colour = "red"
    if toolpath.max_starting_radius >= toolpath.starting_radius:
        colour = "green"

    plt.plot(x, y, linestyle=':', c=colour, linewidth=2)

    if toolpath.max_starting_radius < toolpath.starting_radius:
        # Requested entry circle doesn't fit in the part.
        # Instead max_starting_radius contains the best fit radius.
        starting_circle = geometry.create_circle(
                toolpath.start_point, toolpath.max_starting_radius).path
        x, y = starting_circle.xy
        plt.plot(x, y, linestyle='--', c="red", linewidth=1)

    if toolpath.starting_angle is not None:
        # This implies the entry circle overlaps the cutting path.
        # The contained angle is the direction to the start of the cutting path
        # where it is intersected by the entry circle.
        radius = min(toolpath.starting_radius, toolpath.max_starting_radius)
        line = LineString([
            toolpath.start_point,
            [toolpath.start_point.x, toolpath.start_point.y + radius]
            ])
        line = rotate(line, -toolpath.starting_angle, origin=toolpath.start_point, use_radians=True)

        x, y = line.xy
        plt.plot(x, y, linestyle='--', c="green", linewidth=2)

def generate_tool_path(shapes, step_size, inner=True):
    """ Calculate the toolpath. """

    if inner:
        # Create a polygon that represents an area that does not need cut.
        already_cut = shapes.geoms[0].buffer(-8)

        toolpath = geometry.Pocket(
                shapes,
                step_size,
                geometry.ArcDir.Closest,
                #geometry.ArcDir.CW,
                already_cut=already_cut,
                generate=True,
                #starting_point=Point(-23, 20),
                starting_radius=3.5,
                debug=True)
    else:
        already_cut = Polygon()
        all_pockets = MultiPolygon()
        for shape in shapes.geoms:
            filled_shape = Polygon(shape.exterior)
            desired_pocket = Polygon(filled_shape.exterior).buffer(6).difference(filled_shape)
            cut = desired_pocket.difference(filled_shape.buffer(4))
            already_cut = already_cut.union(cut)
            all_pockets = all_pockets.union(desired_pocket)

        toolpath = geometry.Pocket(
                all_pockets,
                step_size,
                geometry.ArcDir.Closest,
                already_cut=already_cut,
                generate=True,
                starting_point_tactic = geometry.StartPointTactic.PERIMETER,
                #starting_point=Point(0, 42.5),
                starting_radius=0.5,
                debug=True)

    display_outline(already_cut)

    timeslice = 1000  # ms
    for index, progress in enumerate(toolpath.get_arcs(timeslice)):
        print(index, round(progress * 1000) / 1000)
        # toolpath.path contains the currently generated path data at this point.
    return toolpath

def main(argv):
    """
    Example program making use of HSM "peeling" CAM.
    """
    if len(argv) < 2:
        print("Incorrect command line arguments.")
        print(f"Use:\n   {argv[0]} FILENAME [STEP_SIZE]")
        sys.exit(0)
    filename = argv[1]

    try:
        dxf_data = ezdxf.readfile(filename)
    except IOError:
        print(f'Not a DXF file or a generic I/O error.')
        sys.exit(2)
    except ezdxf.DXFStructureError:
        print(f'Invalid or corrupted DXF file.')
        sys.exit(3)

    if len(argv) > 2:
        step_size = float(argv[2])
    else:
        step_size = 1

    print(f"filename: {filename}\n step_size: {step_size}\n")

    modelspace = dxf_data.modelspace()
    shapes = dxf.dxf_to_polygon(modelspace)

    #toolpath = generate_tool_path(shapes, step_size, inner=True)
    toolpath = generate_tool_path(shapes, step_size, inner=False)

    display_outline(shapes)
    display_toolpath(toolpath)
    display_entry_point(toolpath)
    #display_voronoi(toolpath)
    #display_visited_voronoi_edges(toolpath)
    display_start_point(toolpath)

    plt.gca().set_aspect('equal')
    plt.show()

if __name__ == "__main__":
    main(sys.argv)



#!/usr/bin/env python3
"""
Debugging functions.

Recognises the following environment variables:
    HSM_DEBUG: Set to enable any debug output. Default = not set.
    HSM_DEBUG_RES: (Optional) Resolution of output image in dpi. Default = 1000 dpi.
    HSM_DEBUG_FILENAME: (Optional) Specify output image filename. Default = /tmp/hsm.png
    HSM_DEBUG_SCREEN: (Optional) Attempt to output image to screen. Default = not set.
"""

import os

from typing import List, Optional, Union, Any

try:
    import matplotlib.pyplot as plt    # type: ignore
except ImportError:
    pass

from shapely.geometry import MultiPolygon, Polygon  # type: ignore

from hsm_nibble.voronoi_centers import VoronoiCenters  # type: ignore

class Display:
    """
    Use matplotlib.pyplot to display outline of polygons and voronoi diagram.
    """
    filename: str = "/tmp/hsm.png"
    resolution: int = 1000
    screen: bool = False
    initialised: bool = False
    count: int = 0
    colours: List[str] = ["blue", "orange", "green", "cyan", "magenta", "purple", "brown", "grey", "olive", "pink", "yellow"]

    def __init__(self):
        if not os.environ.get("HSM_DEBUG"):
            return

        if not plt:
            print("Error: matplotlib.pyplot not imported.")
            return

        if os.environ.get("HSM_DEBUG_RES"):
            self.resolution = int(os.environ.get("HSM_DEBUG_RES"))

        if os.environ.get("HSM_DEBUG_FILENAME"):
            self.filename = os.environ.get("HSM_DEBUG_FILENAME")

        if os.environ.get("HSM_DEBUG_SCREEN"):
            self.screen = True

        print(f"Writing debug image: {self.filename} at resolution: {self.resolution} dpi")

        plt.gca().set_aspect('equal')

        self.initialised = True

    def display(self,
            polygons: Optional[List[Union[Polygon, MultiPolygon]]] = None,
            voronoi: Optional[VoronoiCenters] = None,
            path: Optional[List] = None) -> None:
        if not self.initialised:
            return

        self.count += 1
        if self.count != 1:
            return
        print(self.count)

        if polygons is None:
            polygons = []

        for index, key in enumerate(polygons.keys()):
            multi = polygons[key]
            colour = self.colours[index % len(self.colours)]
            print(f"  {key}: {colour}")
            if multi.type != "MultiPolygon":
                multi = MultiPolygon([multi])
            for polygon in multi.geoms:
                for ring in [polygon.exterior] + list(polygon.interiors):
                    x, y = ring.xy
                    plt.plot(x, y, c=colour, linewidth=0.01)

        if voronoi:
            for edge in voronoi.edges.values():
                x = []
                y = []
                for point in edge.coords:
                    x.append(point[0])
                    y.append(point[1])
                plt.plot(x, y, c="red", linewidth=0.1)
                plt.plot(x[0], y[0], 'o', c='red', markersize=0.1)
                plt.plot(x[-1], y[-1], 'o', c='red', markersize=0.1)

        if path:
            for entity in path:
                x = []
                y = []
                for point in entity.path.coords:
                    x.append(point[0])
                    y.append(point[1])
                style = "solid"
                colour = "black"
                if type(entity).__name__ == "Line":
                    if str(entity.move_style) == "MoveStyle.RAPID_OUTSIDE":
                        colour = "purple"
                        style = "dotted"
                    if str(entity.move_style) == "MoveStyle.RAPID_INSIDE":
                        colour = "red"
                plt.plot(x, y, c=colour, linestyle=style, linewidth=0.01)

        if self.filename:
            plt.savefig(self.filename, dpi=self.resolution, bbox_inches='tight')
        if self.screen:
            plt.show()

        print("Done writing debug image section.")



#!/usr/bin/env python3

"""
Experimenting with CNC machining toolpaths.
Run code against .dxf test patterns.
"""

from typing import List, NamedTuple, Optional

import argparse
import os
from glob import glob
import signal
import sys
import time

import numpy as np
import ezdxf
import matplotlib.pyplot as plt  # type: ignore
import matplotlib.patches as patches  # type: ignore
from shapely.geometry import box, LineString, MultiLineString, MultiPolygon, Polygon, Point  # type: ignore
from shapely.ops import linemerge, unary_union  # type: ignore
from shapely.errors import GEOSException
from tabulate import tabulate

sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
from hsm_nibble import dxf
from hsm_nibble import geometry

#import warnings
#from shapely.errors import ShapelyDeprecationWarning
#warnings.filterwarnings("error", category=ShapelyDeprecationWarning)


break_count: int = 0
results = []


Result = NamedTuple("Result", [
    ("filename", str),
    ("overlap", float),
    ("winding", geometry.ArcDir),
    ("cut_ratio", float),
    ("crash_ratio", float),
    ("dangerous_crash_count", int),
    ("disjointed_path_count", int),
])

Error = NamedTuple("Error", [
    ("filename", str),
    ("overlap", float),
    ("winding", geometry.ArcDir),
    ("message", str),
])


def signal_handler(_, __):
    """
    Count number of times ^C has bee pressed.
    Exit gracefully after the 1st.
    Actually break after the 2nd.
    """
    global break_count
    global results

    break_count += 1
    if break_count >= 2:
        print(tabulate(results, headers="keys"))
        sys.exit(0)
    print('Ctrl+C pressed. Finishing existing test. Ctrl+C again to quit immediately.')


def init_argparse() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        #usage="%(prog)s [OPTION] [FILE]...",
        description="A program to exercise the CAM HSM 'peeling' algorithm."
    )
    parser.add_argument(
        "-a", "--showarcs",
        action='store_true',
        help="Show the path the endmill takes.")
    parser.add_argument(
        "-f", "--tofile",
        nargs='?',
        const="/tmp/HSM",
        metavar="PATH",
        help="Save a .png of the cut area in the specified directory. "
        "If no directory specified, '/tmp/HSM/' will be used."
    )
    parser.add_argument(
        "-s", "--toscreen",
        action='store_true',
        help="Display cut area on screen")
    parser.add_argument(
        'input_paths',
        nargs='*',
        default=["./test_cases/*.dxf", "../test_cases/*.dxf"],
        metavar="INPUT_PATH",
        help="Path to source .dxf files. "
        "If not specified, will try to find the 'test_cases' directory.")
    return parser


def draw(
        combined_path: List[LineString],
        combined_rapid_inside: List[LineString],
        combined_rapids: List[LineString],
        pocket: Polygon,
        cut_area: Polygon,
        crash_area: Polygon,
        filename: str,
        overlap: float,
        winding: geometry.ArcDir,
        output_path: Optional[str],
        output_display: bool
) -> None:
    """Display cut path."""
    fig, ax = plt.subplots(1, 1, dpi=1600)

    def polygon(poly: Polygon, colour: str, fill: bool = True, linewidth=1) -> None:
        """Draw a Shaplely Polygon on matplotlib.pyplot. """
        if poly.type != "MultiPolygon":
            if poly.type != "Polygon":
                return
            if len(poly.exterior.coords) < 3:
                return
            for interior in poly.interiors:
                if len(interior.coords) < 3:
                    return
            poly = MultiPolygon([poly])
        for geom in poly.geoms:
            if len(geom.exterior.coords) >= 3:
                if fill:
                    patch = patches.Polygon(
                        np.array(geom.exterior.xy).T, fc=colour)
                    ax.add_patch(patch)
                else:
                    ax.plot(*geom.exterior.xy, c=colour, linewidth=linewidth)
                for interior in geom.interiors:
                    if len(interior.coords) >= 3:
                        if fill:
                            patch = patches.Polygon(
                                np.array(interior.xy).T, fc="white")
                            ax.add_patch(patch)
                        else:
                            ax.plot(*interior.xy, c=colour, linewidth=linewidth)

    polygon(cut_area, "green")
    polygon(pocket, "blue", fill=False, linewidth=0.1)
    polygon(crash_area, "red")

    for path in combined_path:
        ax.plot(*path.xy, c="black", linewidth=0.01)
    for path in combined_rapid_inside:
        ax.plot(*path.xy, c="purple", linewidth=0.05, linestyle=(0, (5, 10)))
    for path in combined_rapids:
        ax.plot(*path.xy, c="red", linewidth=0.03)

    ax.axis('equal')
    if output_display:
        plt.show()

    if output_path:
        try:
            os.mkdir(output_path)
        except FileExistsError:
            pass
        except OSError:
            print(f"Could not create directory: {output_path}")
            return

        new_filename = os.path.splitext(filename)[0]
        new_filename = f"{new_filename}_{overlap}_{winding.name}.png"

        plt.savefig(os.path.join(output_path, new_filename))

    plt.cla()
    plt.clf()
    plt.close(fig)
    plt.close("all")

def test_file(
        filepath: str,
        overlap: float,
        winding: geometry.ArcDir,
        show_arcs: bool,
        output_path: Optional[str],
        output_display: bool
) -> Result:
    """Run HSM algorithm on test file. Look for inconsistencies."""
    print(f"Trying: {filepath=}\t{overlap=}\t{winding=}")

    filename = filepath.split("/")[-1]

    dxf_data = ezdxf.readfile(filepath)
    modelspace = dxf_data.modelspace()

    shapes = dxf.dxf_to_polygon(modelspace)
    if shapes.is_valid == False:
        shapes = shapes.buffer(0)
    if isinstance(shapes, Polygon):
        shapes = MultiPolygon([shapes])

    material_bounds = shapes.buffer(20 * overlap).bounds
    material = box(*material_bounds)

    already_cut = material
    for shape in shapes.geoms:
        already_cut = already_cut.difference(Polygon(shape.exterior))

    time_run = time.time()
    toolpath = geometry.Pocket(
            shapes,
            overlap,
            winding,
            already_cut=already_cut,
            generate=False,
            starting_point_tactic = geometry.StartPointTactic.PERIMETER,
            debug=True)
    time_run -= time.time()

    cut_area = Polygon(already_cut)
    newly_cut_area = Polygon()

    combined_path = []
    combined_rapid_inside = []
    combined_rapids = []

    cut_area_parts = []
    crash_area_parts = []

    disjointed_path_count = 0
    last_element = None
    for element in toolpath.path:
        if last_element is not None:
            if (not last_element.end.equals_exact(element.start, 6) or
                 not Point(last_element.path.coords[-1]).equals_exact(
                     Point(element.path.coords[0]), 6)):
                disjointed_path_count += 1
        last_element = element

        if type(element).__name__ == "Arc":
            new_cut = element.path.buffer(overlap / 2)
            cut_area = cut_area.union(new_cut)
            newly_cut_area = newly_cut_area.union(new_cut)

            if show_arcs:
                combined_path.append(element.path)

        elif type(element).__name__ == "Line":
            if element.move_style == geometry.MoveStyle.RAPID_INSIDE:
                new_cut = element.path.buffer(overlap / 2)
                crash = new_cut.difference(cut_area)
                crash_area_parts.append(crash)
                if show_arcs:
                    combined_rapid_inside.append(element.path)
            elif element.move_style == geometry.MoveStyle.RAPID_OUTSIDE:
                if show_arcs:
                    combined_rapids.append(element.path)
            elif element.move_style == geometry.MoveStyle.CUT:
                if show_arcs:
                    combined_path.append(element.path)

    crash_area = unary_union(crash_area_parts)

    uncut_area = toolpath.polygon.difference(cut_area)

    cut_ratio = round(1.0 - uncut_area.area / toolpath.polygon.area, 4)
    crash_ratio = round(crash_area.area / toolpath.polygon.area, 4)

    try:
        eroded_crash_area = crash_area.buffer(-overlap / 20)
    except GEOSException:
        eroded_crash_area = Polygon()
    if eroded_crash_area.type == "Polygon":
        eroded_crash_area = MultiPolygon([eroded_crash_area])
    dangerous_crash_count = len(eroded_crash_area.geoms)

    if output_path or output_display:
        draw(
            combined_path,
            combined_rapid_inside,
            combined_rapids,
            toolpath.polygon,
            newly_cut_area,
            crash_area,
            filename,
            overlap,
            winding,
            output_path,
            output_display)

    return Result(
        filename,
        overlap,
        winding,
        cut_ratio,
        crash_ratio,
        dangerous_crash_count,
        disjointed_path_count,
    )


def main():
    """
    A program to exercise the CAM HSM 'peeling' algorithm.
    Run with --help parameter for usage info.
    """
    global results
    errors = []

    signal.signal(signal.SIGINT, signal_handler)

    parser = init_argparse()
    args = parser.parse_args()

    show_arcs = args.showarcs
    input_paths = args.input_paths
    output_path = args.tofile
    output_display = args.toscreen

    filepaths = []
    for path in input_paths:
        filepaths += glob(path)
    windings = [geometry.ArcDir.CW, geometry.ArcDir.CCW]
    #windings = [geometry.ArcDir.CW, ]
    overlaps = [0.4, 0.8, 1.6, 3.2, 6.4]
    #overlaps = [1.6, 3.2, 6.4]
    #overlaps = [0.4, 1.6]
    count = 0
    total_count = len(filepaths) * len(windings) * len(overlaps)
    for filepath in filepaths:
        for overlap in overlaps:
            for winding in windings:
                count += 1
                try:
                    results.append(test_file(
                        filepath, overlap, winding, show_arcs, output_path, output_display))
                    print(f"{count} of {total_count}")
                    print(results[-1])
                    print()
                except Exception as error:
                    print(error)
                    print(f"during: {filepath}\t{overlap}\t{winding}")
                    errors.append(Error(filepath, overlap, winding, error))
                    raise error

                if break_count:
                    break
            if break_count:
                break
        if break_count:
            break

    print(tabulate(results, headers="keys"))
    print(tabulate(errors, headers="keys"))
    return 0


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
"""
Experimenting with CNC machining toolpaths.
This program is a demo which uses the main library file on test .dxf CAD files.

The demo shows how to calculate a HSM path on the inside of the shape to be cut.
"""

from typing import List, Tuple

import os
import sys
import math
import ezdxf
import matplotlib.pyplot as plt    # type: ignore
from shapely.affinity import rotate  # type: ignore
from shapely.geometry import Point, LineString  # type: ignore

# This line is required if you want to use the local version of the code.
# If you have installed HSM_nibble via PIP it is not required.
sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.abspath(__file__))))

from hsm_nibble import dxf
from hsm_nibble import geometry

#import warnings
#from shapely.errors import ShapelyDeprecationWarning
#warnings.filterwarnings("error", category=ShapelyDeprecationWarning)

def print_entity(entity: ezdxf.entities.DXFGraphic, indent: int = 0):
    """ Display some debug information about a DXF file. """
    dxf_attributes = ["start", "end", "center", "radius", "count"]
    collection_attributes = ["points"]
    other_attributes = ["virtual_entities"]

    padding = " " * indent
    print(f"{padding}{entity}")
    print(f"{padding}  type: {entity.dxftype()}")

    for attribute in dxf_attributes:
        if hasattr(entity.dxf, attribute):
            print(f"{padding}  {attribute}: {getattr(entity.dxf, attribute)}")

    for attribute in collection_attributes:
        if hasattr(entity, attribute):
            generator = getattr(entity, attribute)
            with generator() as collection:
                print(f"{padding}  {attribute}: {collection}")

    for attribute in other_attributes:
        if hasattr(entity, attribute):
            got = getattr(entity, attribute)
            print(f"{padding}  {attribute}: {list(got())}")

def display_outline(shape, colour="blue"):
    """ Display the outline of the shape to be cut. """
    x, y = shape.exterior.xy
    plt.plot(x, y, c=colour, linewidth=2)

    for interior in shape.interiors:
        x, y = interior.xy
        plt.plot(x, y, c=colour, linewidth=2)

def display_voronoi(toolpath, colour="red"):
    """ Display the voronoi edges. These are equidistant from the shape's edges. """
    for edge in toolpath.voronoi.edges.values():
        x = []
        y = []
        for point in edge.coords:
            x.append(point[0])
            y.append(point[1])
        plt.plot(x, y, c="red", linewidth=4)
        plt.plot(x[0], y[0], 'x', c=colour)
        plt.plot(x[-1], y[-1], 'x', c=colour)

def display_visited_voronoi_edges(toolpath, colour="black"):
    """
    Display the voronoi edges that were used to calculate cut geometry.
    This should match the output of display_voronoi(...).
    """
    for edge in toolpath.visited_edges:
        edge_coords = toolpath.voronoi.edges[edge].coords
        x = []
        y = []
        for point in edge_coords:
            x.append(point[0])
            y.append(point[1])
        plt.plot(x, y, c=colour, linewidth=1)
        plt.plot(x[0], y[0], 'x', c=colour)
        plt.plot(x[-1], y[-1], 'x', c=colour)

def display_entry_point(toolpath):
    if toolpath.starting_radius is None:
        # Entry circle not set.
        return

    starting_circle = geometry.create_circle(
            toolpath.start_point, toolpath.starting_radius).path
    x, y = starting_circle.xy

    colour = "red"
    if toolpath.max_starting_radius >= toolpath.starting_radius:
        colour = "green"

    plt.plot(x, y, linestyle=':', c=colour, linewidth=2)

    if toolpath.max_starting_radius < toolpath.starting_radius:
        # Requested entry circle doesn't fit in the part.
        # Instead max_starting_radius contains the best fit radius.
        starting_circle = geometry.create_circle(
                toolpath.start_point, toolpath.max_starting_radius).path
        x, y = starting_circle.xy
        plt.plot(x, y, linestyle='--', c="red", linewidth=1)

    if toolpath.starting_angle is not None:
        # This implies the entry circle overlaps the cutting path.
        # The contained angle is the direction to the start of the cutting path
        # where it is intersected by the entry circle.
        radius = min(toolpath.starting_radius, toolpath.max_starting_radius)
        line = LineString([
            toolpath.start_point,
            [toolpath.start_point.x, toolpath.start_point.y + radius]
            ])
        line = rotate(line, -toolpath.starting_angle, origin=toolpath.start_point, use_radians=True)

        x, y = line.xy
        plt.plot(x, y, linestyle='--', c="green", linewidth=2)

def display_starting_circle(toolpath, colour="orange"):
    starting_circle = geometry.create_circle(
            toolpath.start_point, toolpath.start_radius).path
    x, y = starting_circle.xy
    plt.plot(x, y, c=colour, linewidth=4)

def display_toolpath(toolpath, cut_colour="green", rapid_inside_colour="blue", rapid_outside_colour="orange"):
    # Display path.
    last_element = None

    for element in toolpath.path:
        if last_element is not None:
            assert last_element.end.equals_exact(element.start, 6)
            assert Point(last_element.path.coords[-1]
                    ).equals_exact(Point(element.path.coords[0]), 6)
        last_element = element

        if type(element).__name__ == "Arc":
            x, y = element.path.xy
            if element.debug:
                plt.plot(x, y, c=element.debug, linewidth=3)
            else:
                plt.plot(x, y, c=cut_colour, linewidth=1)

        elif type(element).__name__ == "Line":
            x, y = element.path.xy
            if element.move_style == geometry.MoveStyle.RAPID_INSIDE:
                plt.plot(x, y, linestyle='--', c=rapid_inside_colour, linewidth=1)
            elif element.move_style == geometry.MoveStyle.RAPID_OUTSIDE:
                plt.plot(x, y, c=rapid_outside_colour, linewidth=1)
            else:
                assert element.move_style == geometry.MoveStyle.CUT
                plt.plot(x, y, linestyle='--', c=cut_colour, linewidth=1)


def generate_tool_path(shape, step_size):
    """ Calculate the toolpath. """
    toolpath = geometry.Pocket(
            shape,
            step_size,
            #geometry.ArcDir.Closest,
            geometry.ArcDir.CW,
            #geometry.ArcDir.CCW,
            generate=True,
            #starting_point=Point(-39.9, 11.8),
            #starting_radius=2.5,
            debug=True)

    timeslice = 1000  # ms
    for index, progress in enumerate(toolpath.get_arcs(timeslice)):
        print(index, round(progress * 1000) / 1000)
        # toolpath.path contains the currently generated path data at this point.
    return toolpath

def main(argv):
    """
    Example program making use of HSM "peeling" CAM.
    """
    if len(argv) < 2:
        print("Incorrect command line arguments.")
        print(f"Use:\n   {argv[0]} FILENAME [STEP_SIZE]")
        sys.exit(0)
    filename = argv[1]

    try:
        dxf_data = ezdxf.readfile(filename)
    except IOError:
        print(f'Not a DXF file or a generic I/O error.')
        sys.exit(2)
    except ezdxf.DXFStructureError:
        print(f'Invalid or corrupted DXF file.')
        sys.exit(3)

    if len(argv) > 2:
        step_size = float(argv[2])
    else:
        step_size = 1

    print(f"filename: {filename}\n step_size: {step_size}\n")

    modelspace = dxf_data.modelspace()
    shape = dxf.dxf_to_polygon(modelspace).geoms[-1]
    #shape = dxf.dxf_to_polygon(modelspace).geoms[0]

    toolpath = generate_tool_path(shape, step_size)
    # Call toolpath.calculate_path() to scrap the existing and regenerate toolpath.

    display_outline(shape)
    display_entry_point(toolpath)
    #display_starting_circle(toolpath)
    display_toolpath(toolpath)
    display_voronoi(toolpath)
    # display_visited_voronoi_edges(toolpath)

    plt.gca().set_aspect('equal')
    plt.show()

if __name__ == "__main__":
    main(sys.argv)



#!/usr/bin/env python3

import unittest
import os, sys
import math

from shapely.geometry import LineString, Polygon  # type: ignore

sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.abspath(__file__))))

import hsm_nibble.voronoi_centers as voronoi_centers

class TestBoilerplate(unittest.TestCase):
    def setUp(self):
        self.vc = voronoi_centers.VoronoiCenters()
        self.vc.vertex_to_edges = {
                (10.1, 10.1): [1, 2],
                (20.2, 10.1): [2, 4],
                (20.2, 20.2): [4, 5, 3],
                (10.1, 20.1): [1, 6, 5],
                (30.3, 20.2): [3,],
                (-10, 20): [6,],
                }
        self.vc.edge_to_vertex = {
                1: ((10.1, 10.1), (10.1, 20.1)),
                2: ((20.2, 10.1), (10.1, 10.1)),
                3: ((30.3, 20.2), (20.2, 20.2)),
                4: ((20.2, 20.2), (20.2, 10.1)),
                5: ((20.2, 20.2), (10.1, 20.1)),
                6: ((10.1, 20.1), (-10, 20)),
                }
        self.vc.edges = {
                1: LineString(self.vc.edge_to_vertex[1]),
                2: LineString(self.vc.edge_to_vertex[2]),
                3: LineString(self.vc.edge_to_vertex[3]),
                4: LineString(self.vc.edge_to_vertex[4]),
                5: LineString(self.vc.edge_to_vertex[5]),
                6: LineString(self.vc.edge_to_vertex[6]),
                }
        self.vc._check_data()

    def test_one(self):
        pass


if __name__ == '__main__':
    unittest.main()




"""
Helper methods for converting .dxf source files into shapely Polygons.
Used by the test implementation main.py.
"""

from typing import List, Tuple

from math import *

from ezdxf.entities.lwpolyline import LWPolyline as DxfPolyline
from ezdxf.query import EntityQuery as DxfPolylines
from shapely.geometry import LineString, MultiLineString, Point, Polygon, LinearRing, MultiPoint, MultiPolygon  # type: ignore

CIRCLE_RES = 500

# Introduce coordinate jitter by clamping coordinates to this number of decimal places.
# Used for testing.
# Set to "0" to disable.
DP = 0

def round_points(points):
    """ Used in testing. Clamps points to rounded coordinates. """
    if not DP:
        return points
    return [(round(point[0], DP), round(point[1], DP)) for point in points]

def weighted(p1, p2, alpha):
    return p1[0] + (p2[0] - p1[0]) * alpha, p1[1] + (p2[1] - p1[1]) * alpha

def circle(x, y, r, n=None, sa=0, ea=2*pi):
    if n is None:
        n = pi * r * GeometrySettings.RESOLUTION
    n *= abs((ea - sa) / (2 * pi))
    n = ceil(n)
    res = []
    for i in range(n + 1):
        a = sa + i * (ea - sa) / n
        newpt = (x + r * cos(a), y + r * sin(a))
        if not res or newpt != res[-1]:
            res.append(newpt)
    return res

def circle_to_linestring(circle) -> LineString:
    radius = getattr(circle.dxf, "radius")
    center = getattr(circle.dxf, "center")
    return Point(center).buffer(radius).simplify(0.1).boundary

def polyline_to_linestring(entity) -> LineString:
    points: List[Tuple[float, float]] = []
    lastx, lasty = entity[-1][0:2]
    lastbulge = entity[-1][4]
    for point in entity:
        x, y = point[0:2]
        if lastbulge:
            theta = 4 * atan(lastbulge)
            dx, dy = x - lastx, y - lasty
            mx, my = weighted((lastx, lasty), (x, y), 0.5)
            angle = atan2(dy, dx)
            dist = sqrt(dx * dx + dy * dy)
            d = dist / 2
            r = abs(d / sin(theta / 2))
            c = d / tan(theta / 2)
            cx = mx - c * sin(angle)
            cy = my + c * cos(angle)
            sa = atan2(lasty - cy, lastx - cx)
            ea = sa + theta
            points += round_points(circle(cx, cy, r, CIRCLE_RES, sa, ea))
            points.append((x, y))
        else:
            points.append((x, y))
        lastbulge = point[4]
        lastx, lasty = x, y
    return LineString(points)

def dxf_to_polygon(modelspace) -> MultiPolygon:
    """ Convert .dxf polyline into shapely Polygon. """
    rings = []
    for entity in modelspace:
        if entity.dxftype() == "CIRCLE":
            rings.append(circle_to_linestring(entity))
        elif entity.dxftype() == "LWPOLYLINE":
            rings.append(polyline_to_linestring(entity))
        else:
            print(f"Unsupported dxf entity: {entity.dxftype()}")

    parents = []
    for ring in rings:
        coveres = []
        covered_by = False
        for parent_index, parent in enumerate(parents):
            if Polygon(ring).covers(parent):
                coveres.append(parent_index)
            if Polygon(parent).covers(ring):
                covered_by = True
        for not_parent in sorted(coveres, reverse=True):
            parents.pop(not_parent)
        if not covered_by:
            parents.append(ring)

    polygons = []
    for parent in parents:
        holes = []
        for ring in rings:
            if ring is parent:
                continue
            if Polygon(parent).covers(ring):
                holes.append(ring)
        polygon = Polygon(parent, holes=holes)
        if not polygon.is_valid:
            polygon = polygon.buffer(0)
        polygons.append(polygon)

    return MultiPolygon(polygons)



#!/usr/bin/env python3
"""
Experimenting with CNC machining toolpaths.
Run code against .dxf test patterns.
"""

from typing import NamedTuple, Tuple

from glob import glob
import os
import signal
import sys
from tabulate import tabulate
import time

import ezdxf
from shapely.geometry import Polygon  # type: ignore

sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
from hsm_nibble import dxf
from hsm_nibble import geometry


break_count: int = 0

Result = NamedTuple("Result", [
    ("filename", str),
    ("overlap", float),
    ("winding", geometry.ArcDir),
    ("uncut_area", float),
    ("uncut_ratio", float),
    ("arc_count", int),
    ("arc_attempt_ratio", float),
    ("arc_fail_ratio", float),
    ("path_fail_ratio", float),
    ("time_per_arc", float),
    ])

Error = NamedTuple("Error", [
    ("filename", str),
    ("overlap", float),
    ("winding", geometry.ArcDir),
    ("error_message", str),
    ])

def signal_handler(sig, frame):
    global break_count
    break_count += 1
    if break_count >= 2:
        sys.exit(0)
    print('Ctrl+C pressed. Finishing existing test. Ctrl+C again to quit immediately.')

def describe():
    print("uncut_area:")
    print("\tThe area enclosed by the drawn path is subtracted from the "
            "desired geometry shape. The remainder is then eroded by the overlap / 2. "
            "(Lower is better)")
    print("uncut_ratio:")
    print("\tThe uncut_area divided by the number of arcs to give a "
            "normalized ratio. (Lower is better)")
    print("arc_count:")
    print("\tThe number of arcs in the calculated path.")
    print("arc_attempt_ratio:")
    print("\tThe average number of arc calculations before arriving "
            "as an arc that fits. (Lower is better.)")
    print("arc_fail_ratio:")
    print("\tThe ratio of failed attempts to find an exact fit arc. "
            "(Lower is better)")
    print("path_fail_ratio:")
    print("\tThe number of failures to draw any more arcs along a "
            "voronoi edge. Any number over 0 should be investigated. (Lower is better)")
    print("time_per_arc:")
    print("\tThe average time taken to calculate an arc position. "
            "(Lower is better)")
    print()

def test_file(filepath: str, overlap: float, winding: geometry.ArcDir) -> Result:
    print(f"Trying: {filepath=}\t{overlap=}\t{winding=}")

    filename = filepath.split("/")[-1]

    dxf_data = ezdxf.readfile(filepath)
    modelspace = dxf_data.modelspace()
    shape = dxf.dxf_to_polygon(modelspace)

    time_run = time.time()
    toolpath = geometry.Pocket(shape, overlap, winding, generate=False, debug=True)
    time_run -= time.time()
    
    polygon_remaining = toolpath.polygon
    #center_circle = toolpath.start_point.buffer(toolpath.start_radius)
    #polygon_remaining = polygon_remaining.difference(center_circle)
    for element in toolpath.path:
        if type(element).__name__ == "Arc":
            polygon_remaining = polygon_remaining.difference(
                    element.path.buffer(overlap))

    polygon_area = round(toolpath.polygon.area, 4)
    uncut_area = round(polygon_remaining.area, 4)
    uncut_ratio = round(polygon_remaining.area/polygon_area, 4)

    arc_count = len(toolpath.path)
    if arc_count:
        arc_attempt_ratio = round(toolpath.loop_count / arc_count, 4)
        arc_fail_ratio = round(toolpath.arc_fail_count / arc_count, 4)
        path_fail_ratio = toolpath.path_fail_count / arc_count
        time_per_arc = round(-time_run / arc_count, 4)
    else:
        arc_attempt_ratio = "infinite"
        arc_fail_ratio = "infinite"
        path_fail_ratio = "infinite"
        time_per_arc = "infinite"

    return Result(
            filename,
            overlap,
            winding,
            uncut_area,
            uncut_ratio,
            arc_count,
            arc_attempt_ratio,
            arc_fail_ratio,
            path_fail_ratio,
            time_per_arc,
            )

def help(progname: str):
    print("A program to exercise the CAM HSM 'peeling' algorithm.\n\n"
            "It uses the library to generate CAM paths for each .dxf CAD test "
            "file to gather statistics.\n"
            "It will search the nearby path for test cases is no path is specified.\n\n"
            "Usage:\n"
            "  {progname} [dxf_files_glob]\n"
            "eg:\n"
            "  {progname}\n"
            "  {progname} ./some/dir/\*.dxf\n"
            "  {progname} ./test_cases/curves\*.dxf\n".format(progname=progname))

def main(argv):
    """
    A program to exercise the CAM HSM 'peeling' algorithm.
    Run with --help parameter for usage info.
    """
    signal.signal(signal.SIGINT, signal_handler)

    if {"-h", "--h", "-help", "--help", "help"} & set(argv):
        help(argv[0])
        return 0

    paths = ["../test_cases/*.dxf", "./test_cases/*.dxf"]
    if len(argv) >= 2:
        paths = [argv[1]]

    results = []
    failures = []

    filepaths = []
    for path in paths:
        filepaths += glob(path)
    windings = [geometry.ArcDir.CW, geometry.ArcDir.CCW, geometry.ArcDir.Closest]
    #windings = [geometry.ArcDir.CW,]
    overlaps = [0.4, 0.8, 1.6, 3.2, 6.4]
    #overlaps = [0.4, 1.6]
    count = 0
    total_count = len(filepaths) * len(windings) * len(overlaps)
    for filepath in filepaths:
        for overlap in overlaps:
            for winding in windings:
                count += 1
                try:
                    results.append(test_file(filepath, overlap, winding))
                    print(f"{count} of {total_count}")
                    print(results[-1])
                    print()
                except Exception as error:
                    print(error)
                    print(f"during: {filepath}\t{overlap}\t{winding}")
                    failures.append(Error(filepath.split("/")[-1], overlap, winding, error,))
                    #raise error

                if break_count:
                    break
            if break_count:
                break
        if break_count:
            break
    describe()
    print(tabulate(results, headers="keys"))
    print(tabulate(failures, headers="keys"))

if __name__ == "__main__":
    main(sys.argv)




#!/usr/bin/env python3
"""
Experimenting with CNC machining toolpaths.
This program is a demo which uses the main library file on test .dxf CAD files.

The demo shows how to calculate a HSM path on the outside of the shape to be cut.
"geometry.Pocket(...)" takes an "already_cut" parameter which represents the outer boundary of the
area to be cut. eg: The edge of the stock size.
"""

import os
import sys

import ezdxf
import matplotlib.pyplot as plt    # type: ignore
from shapely.geometry import box, LinearRing, LineString, MultiPolygon, Point, Polygon  # type: ignore

# This line is required if you want to use the local version of the code.
# If you have installed HSM_nibble via PIP it is not required.
sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.abspath(__file__))))

from hsm_nibble import dxf
from hsm_nibble import geometry


def display_outline(shape, colour="blue"):
    """ Display the outline of the shape to be cut. """
    shapes = shape
    if shapes.type != "MultiPolygon":
        shapes = MultiPolygon([shapes])

    for shape in shapes.geoms:
        x, y = shape.exterior.xy
        plt.plot(x, y, c=colour, linewidth=2)

        for interior in shape.interiors:
            x, y = interior.xy
            plt.plot(x, y, c=colour, linewidth=2)

def display_voronoi(toolpath, colour="red"):
    """ Display the voronoi edges. These are equidistant from the shape's edges. """
    for edge in toolpath.voronoi.edges.values():
        x = []
        y = []
        for point in edge.coords:
            x.append(point[0])
            y.append(point[1])
        plt.plot(x, y, c="red", linewidth=4)
        plt.plot(x[0], y[0], 'x', c=colour)
        plt.plot(x[-1], y[-1], 'x', c=colour)

def display_visited_voronoi_edges(toolpath, colour="black"):
    """
    Display the voronoi edges that were used to calculate cut geometry.
    This should match the output of display_voronoi(...).
    """
    for edge in toolpath.visited_edges:
        edge_coords = toolpath.voronoi.edges[edge].coords
        x = []
        y = []
        for point in edge_coords:
            x.append(point[0])
            y.append(point[1])
        plt.plot(x, y, c=colour, linewidth=1)
        plt.plot(x[0], y[0], 'x', c=colour)
        plt.plot(x[-1], y[-1], 'x', c=colour)

def display_toolpath(toolpath, cut_colour="green", rapid_inside_colour="blue", rapid_outside_colour="orange"):
    # Display path.
    last_element = None
    for element in toolpath.path:
        if last_element is not None:
            assert last_element.end.equals_exact(element.start, 6)
            assert Point(last_element.path.coords[-1]
                    ).equals_exact(Point(element.path.coords[0]), 6)
        last_element = element

        if type(element).__name__ == "Arc":
            x, y = element.path.xy
            if element.debug:
                plt.plot(x, y, c=element.debug, linewidth=3)
            else:
                plt.plot(x, y, c=cut_colour, linewidth=1)

        elif type(element).__name__ == "Line":
            x, y = element.path.xy
            if element.move_style == geometry.MoveStyle.RAPID_INSIDE:
                plt.plot(x, y, linestyle='--', c=rapid_inside_colour, linewidth=1)
            elif element.move_style == geometry.MoveStyle.RAPID_OUTSIDE:
                plt.plot(x, y, c=rapid_outside_colour, linewidth=1)
            else:
                assert element.move_style == geometry.MoveStyle.CUT
                plt.plot(x, y, linestyle='--', c=cut_colour, linewidth=1)


def display_start_point(toolpath, colour="purple"):
    plt.plot(toolpath.start_point.x, toolpath.start_point.y, marker='o', c=colour, markersize=10)

def main(argv):
    """
    Example program making use of HSM "peeling" CAM.
    """
    if len(argv) < 2:
        print("Incorrect command line arguments.")
        print(f"Use:\n   {argv[0]} FILENAME [STEP_SIZE]")
        sys.exit(0)
    filename = argv[1]

    try:
        dxf_data = ezdxf.readfile(filename)
    except IOError:
        print(f'Not a DXF file or a generic I/O error.')
        sys.exit(2)
    except ezdxf.DXFStructureError:
        print(f'Invalid or corrupted DXF file.')
        sys.exit(3)

    if len(argv) > 2:
        step_size = float(argv[2])
    else:
        step_size = 1

    print(f"filename: {filename}\n step_size: {step_size}\n")

    modelspace = dxf_data.modelspace()

    shapes = dxf.dxf_to_polygon(modelspace)
    if shapes.is_valid == False:
        shapes = shapes.buffer(0)

    material_bounds = shapes.buffer(20 * step_size).bounds
    material = box(*material_bounds)

    already_cut = material
    for shape in shapes.geoms:
        already_cut = already_cut.difference(Polygon(shape.exterior))

    toolpath = geometry.Pocket(
            shapes,
            step_size,
            geometry.ArcDir.Closest,
            already_cut=already_cut,
            generate=True,
            starting_point_tactic = geometry.StartPointTactic.PERIMETER,
            debug=True)

    timeslice = 100  # ms
    for index, progress in enumerate(toolpath.get_arcs(timeslice)):
        print(index, progress)
        # toolpath.path contains the currently generated path data at this point.

    #display_outline(shapes)
    display_toolpath(toolpath)
    display_voronoi(toolpath)
    # display_visited_voronoi_edges(toolpath)
    display_start_point(toolpath)

    plt.gca().set_aspect('equal')
    plt.show()

if __name__ == "__main__":
    main(sys.argv)



"""
    A CAM library for generating HSM "peeling" toolpaths from supplied geometry.

    Copyright (C) <2022>  <duncan law>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

    mrdunk@gmail.com
"""
from typing import Dict, List, Optional, Set, Tuple, Union

import math

from shapely.geometry.base import BaseGeometry  # type: ignore
from shapely.geometry import box, LineString, Point, Polygon  # type: ignore
from shapely.ops import linemerge, nearest_points  # type: ignore
from shapely.validation import make_valid  # type: ignore

import pyvoronoi  # type: ignore

from hsm_nibble.helpers import log  # type: ignore

Vertex = Tuple[float, float]

# Number of decimal places resolution for coordinates.
ROUND_DP = 5

# Resolution of voronoi algorithm.
# See C++ Boost documentation.
# Set it to 1 better than the geometry resolution.
VORONOI_RES = 10**(ROUND_DP + 1)

# A small number for comparing things that should touch if not for floating-point error.
#EPS = 5.96e-08
EPS = 1 / (10 ** ROUND_DP)


class PointOutsidePart(Exception):
    pass


def round_coord(value: Tuple[float, float], dp: int = ROUND_DP) -> Tuple[float, float]:
    return (round(value[0], dp), round(value[1], dp))


class VoronoiCenters:
    """
    A wrapper for pyvoronoi that calculates midpoints equidistant to 2 or more polygon
    edges.
    While pyvoronoi does not natively handle arcs and circles, this wrapper attempts
    to prune voronoi edges that are produced by arcs that have been split into line
    segments.
    """

    def __init__(
            self,
            polygon: Optional[Polygon] = None,
            tolerence: float = 1.01,
            preserve_widest: bool = False,
            preserve_edge: bool = False,
            starting_point: Point = None
            ) -> None:
        """
        Arguments:
            polygon: The geometer that we wish to generate a voronoi diagram inside.
            tolerence: Parameter used for pruning unwanted voronoi edges. This should
                be approximately the same as the maximum expected jitter in
                geometry coordinates.
        """
        if not polygon:
            return

        self.polygon = polygon
        self._validate_poly()
        self.tolerence = tolerence

        self.max_dist = max((self.polygon.bounds[2] - self.polygon.bounds[0]),
                            (self.polygon.bounds[3] - self.polygon.bounds[1])) + 1

        # Collect polygon segments to populate voronoi with.
        geom_primatives: List[Tuple[Vertex, Vertex]] = []
        outlines = [self.polygon.exterior] + list(self.polygon.interiors)
        for outline in outlines:
            prev_point = None
            first_point = None
            for point in outline.coords:
                if prev_point is None:
                    first_point = point
                    prev_point = point
                elif point == prev_point:
                    continue
                else:
                    geom_primatives.append((prev_point, point))
                    prev_point = point
            assert prev_point == first_point  # This is a loop.

        # Generate voronoi diagram.
        pv = pyvoronoi.Pyvoronoi(VORONOI_RES)
        for segment in geom_primatives:
            pv.AddSegment(segment)
        pv.Construct()

        edges = pv.GetEdges()
        vertices = pv.GetVertices()
        cells = pv.GetCells()

        # Parse voronoi diagram. Store data as shapely LineSegment.
        self.edges: Dict[int, LineString] = {}
        self.vertex_to_edges: Dict[Vertex, List[int]] = {}
        self.edge_to_vertex: Dict[int, Tuple[Vertex, Vertex]] = {}

        self.edge_count = 0
        visited_edges: Set[int] = set()
        for edge_index, edge in enumerate(edges):
            if edge_index in visited_edges:
                continue
            visited_edges.add(edge_index)
            visited_edges.add(edge.twin)

            start_vert = vertices[edge.start]
            end_vert = vertices[edge.end]
            if edge.twin and edge.is_primary:
                p_start = Point(round_coord((start_vert.X, start_vert.Y)))
                p_end = Point(round_coord((end_vert.X, end_vert.Y)))
                if (p_start.distance(self.polygon) <= EPS and
                        p_end.distance(self.polygon) <= EPS):
                    if edge.is_linear:
                        line = LineString((
                            round_coord((start_vert.X, start_vert.Y)),
                            round_coord((end_vert.X, end_vert.Y))
                        ))
                        self._store_edge(line)
                    else:
                        cell = cells[edge.cell]
                        cell_twin = cells[edges[edge.twin].cell]

                        geom_points = []
                        geom_edges = []
                        if cell.contains_point:
                            geom_points.append(pv.RetrieveScaledPoint(cell))
                        if cell_twin.contains_point:
                            geom_points.append(
                                pv.RetrieveScaledPoint(cell_twin))
                        if cell.contains_segment:
                            geom_edges.append(pv.RetriveScaledSegment(cell))
                        if cell_twin.contains_segment:
                            geom_edges.append(
                                pv.RetriveScaledSegment(cell_twin))
                        assert len(geom_edges) > 0
                        assert len(geom_points) + len(geom_edges) == 2

                        if len(geom_points) == 1 and len(geom_edges) == 1:
                            start_point = Point(round_coord(
                                (start_vert.X, start_vert.Y)))
                            end_point = Point(round_coord(
                                (end_vert.X, end_vert.Y)))
                            max_distance = start_point.distance(end_point) / 10 + 0.01
                            points = (round_coord(point)
                                      for point in
                                      pv.DiscretizeCurvedEdge(edge_index, max_distance))
                            self._store_edge(LineString(points))
                        else:
                            # A parabola between 2 lines (as opposed to 1 line and one point)
                            # leaves the DiscretizeCurvedEdge() function broken sometimes.
                            # Let's just assume a straight line edge in these cases.
                            # This is a particular problem when duplicate points in
                            # input data create a line of zero length.
                            log("BORKED VORONOI: \t"
                                "geom_points: {geom_points}\tgeom_edges: {geom_edges}")
                            line = LineString((
                                round_coord((start_vert.X, start_vert.Y)),
                                round_coord((end_vert.X, end_vert.Y))
                            ))
                            self._store_edge(line)
                        continue

        preserve = set()
        if preserve_widest:
            # The widest_gap() method checks vertices for the widest point.
            # If we remove this vertex subsequent calls will not work.
            # If we cached the value instead, client code will not be able to use the
            # returned vertex value as an index into this class's data structures.
            self.start_point, self.start_distance = self.widest_gap()
            if self.start_point is not None:
                preserve.add(self.start_point.coords[0])
        if preserve_edge:
            # The vertex_on_perimiter() method picks a vertex that touches the
            # perimeter. We might not want to clean that up if we want to cut in from
            # the edge.
            self.start_point = self.vertex_on_perimiter()
            preserve.add(self.start_point.coords[0])
            self.start_distance = 0

        self._drop_irrelevant_edges(preserve)

        if starting_point is not None:
            # Set the starting point.
            self.set_starting_point(starting_point)
            preserve.add(self.start_point.coords[0])

        self._split_on_pocket_edge()
        self._combine_edges(preserve)
        self._remove_trivial(preserve)

        self.max_starting_radius = self.distance_from_geom(self.start_point)

        self._check_data()

    def _check_data(self) -> None:
        """ Sanity check data structures. """
        for edge_i, edge in self.edges.items():
            assert edge_i in self.edge_to_vertex
            vertices = self.edge_to_vertex[edge_i]
            assert len(vertices) == 2
            assert edge.coords[0] in vertices
            assert edge.coords[-1] in vertices
            assert edge.coords[0] != edge.coords[-1]  # Loop
            assert vertices[0] in self.vertex_to_edges
            assert vertices[1] in self.vertex_to_edges

        for edge_i, vertices in self.edge_to_vertex.items():
            assert edge_i in self.edges
            assert len(vertices) == 2
            assert edge_i in self.edges
            assert vertices[0] in self.vertex_to_edges
            assert vertices[1] in self.vertex_to_edges

        for vertex, edges_i in self.vertex_to_edges.items():
            assert len(set(edges_i)) == len(edges_i)  # Loop
            for edge_i in edges_i:
                assert edge_i in self.edges
                vertices = self.edge_to_vertex[edge_i]
                assert vertex in vertices

    def _validate_poly(self) -> None:
        """
        Make sure input geometry is sane.
        Do some quick fixes on common issues.
        """
        fixed = make_valid(self.polygon)
        while fixed.geom_type == "MultiPolygon":
            bigest = None
            size = 0
            for geom in fixed.geoms:
                poly_area = Polygon(geom).area
                if poly_area > size:
                    size = poly_area
                    bigest = geom
            fixed = bigest
            fixed = make_valid(fixed)
            # TODO: Should we just throw an exception here?
            # There is more than one choice for the outer geometry loop.
            # Knowing which piece to work on is a crap shoot.
            # It should be up to the client code to make the decision and correctly
            # format the input polygon.
        if fixed.geom_type == "Polygon":
            self.polygon = fixed

        # Any jitter in input geometry may make points appear out of sequence
        # leading to invalid geometry.
        # See here for what valid geometry looks like:
        # https://shapely.readthedocs.io/en/latest/manual.html#polygons
        # .simplify(...) will fix issues caused by input coordinate jitter.
        self.polygon = self.polygon.simplify(0.05)

    def _store_edge(self, edge: LineString, replace_index=None):
        """
        Store a vorinoi edge and associated vertices in out internal data structures.
        """
        if edge.length == 0:
            return

        edge_index = replace_index
        if edge_index is None:
            edge_index = self.edge_count
            self.edge_count += 1
        vert_index_a = (edge.coords[0][0], edge.coords[0][1])
        vert_index_b = (edge.coords[-1][0], edge.coords[-1][1])

        self.edges[edge_index] = edge
        if edge_index not in self.vertex_to_edges:
            self.vertex_to_edges.setdefault(
                vert_index_a, []).append(edge_index)
        if edge_index not in self.vertex_to_edges:
            self.vertex_to_edges.setdefault(
                vert_index_b, []).append(edge_index)
        self.edge_to_vertex[edge_index] = (vert_index_a, vert_index_b)

        return

    def _remove_edge(self, edge_index: int) -> None:
        """
        Remove a vorinoi edge and associated vertices from internal data structures.
        """
        if len(self.edges) <= 1:
            # Special case when cleaning up edges and self.polygon is a simple circle.
            return
        vert_a, vert_b = self.edge_to_vertex[edge_index]
        neibours_a = self.vertex_to_edges[vert_a]
        neibours_b = self.vertex_to_edges[vert_b]

        neibours_a.remove(edge_index)
        neibours_b.remove(edge_index)
        if not neibours_a:
            del self.vertex_to_edges[vert_a]
        if not neibours_b:
            del self.vertex_to_edges[vert_b]
        del self.edge_to_vertex[edge_index]
        del self.edges[edge_index]

    def distance_from_geom(self, point: BaseGeometry) -> float:
        """
        Distance form nearest geometry edge. Note this edge may be the outer
        boundary or the edge of an island.
        """
        distance = self.max_dist
        for ring in [self.polygon.exterior] + list(self.polygon.interiors):
            distance = min(distance, ring.distance(point))
        return distance

    def _combine_edges(self, dont_merge: Set[Tuple[float, float]]) -> None:
        """
        For any voronoi vertex with only 2 edges meeting there, combine the
        voronoi edges, removing the vertex.
        """
        to_merge = []
        for vertex, edges_i in self.vertex_to_edges.items():
            if len(edges_i) == 2 and vertex not in dont_merge:
                to_merge.append(vertex)

        while to_merge:
            candidate = to_merge.pop()
            edges_i = self.vertex_to_edges[candidate]
            assert len(edges_i) == 2

            if edges_i[0] not in self.edges or edges_i[1] not in self.edges:
                # Already done.
                continue

            edge_i_a = edges_i[0]
            edge_i_b = edges_i[1]

            edge_a = self.edges[edge_i_a].coords
            edge_b = self.edges[edge_i_b].coords

            if edge_a[0] in [edge_b[0], edge_b[-1]] and edge_a[-1] in [edge_b[0], edge_b[-1]]:
                # Loop with no branches. "Circle on a stick."
                # It wold be possible to combine the last 2 segments but that
                # would confuse linemerge(...) as it would not know which pair
                # of ends to make the beginning and end of the LineString.
                continue

            self._remove_edge(edge_i_a)
            self._remove_edge(edge_i_b)

            edge_combined = linemerge([edge_a, edge_b])
            self._store_edge(edge_combined, edge_i_a)

    def _remove_trivial(self, preserve: Set[Tuple[float, float]]):
        """Collapse any edge that is too short to care about."""
        # First the easy cases: The short edges with no connections at one end.
        further_consideration = set()
        run = True
        while run:
            run = False
            for edge_i, edge in self.edges.items():
                if edge.length < EPS * 10 and len(self.edges) > 1:
                    vert_a, vert_b = self.edge_to_vertex[edge_i]
                    if len(self.vertex_to_edges[vert_a]) == 1 and vert_a not in preserve:
                        self._remove_edge(edge_i)
                        run = True
                        break
                    if len(self.vertex_to_edges[vert_b]) == 1 and vert_b not in preserve:
                        self._remove_edge(edge_i)
                        run = True
                        break
                    further_consideration.add(edge_i)

        # Some edges are duplicates of each other, joined at both ends.
        # Like tiny loops.
        still_to_remove: Dict[Tuple[Vertex, Vertex], Set[int]] = {}
        for edge_i in further_consideration:
            vert_a, vert_b = self.edge_to_vertex[edge_i]

            # Make a key that is the same for any edge starting and finishing in the same place.
            key = (vert_a, vert_b)
            if vert_a < vert_b:
                key = (vert_b, vert_a)

            still_to_remove.setdefault(key, set()).add(edge_i)
        for edges_i in still_to_remove.values():
            while len(edges_i) > 1 and len(self.edges) > 1:
                edge_i = edges_i.pop()
                self._remove_edge(edge_i)

        # Duplicates are now gone.
        # Clean up simple lines where appropriate.
        for edges_i in still_to_remove.values():
            edge_i = edges_i.pop()
            vert_a, vert_b = self.edge_to_vertex[edge_i]
            if len(self.vertex_to_edges[vert_a]) == 1 and vert_a not in preserve:
                if len(self.edges) > 1:
                    self._remove_edge(edge_i)
            if len(self.vertex_to_edges[vert_b]) == 1 and vert_b not in preserve:
                if len(self.edges) > 1 and edge_i in self.edges:
                    self._remove_edge(edge_i)

    def _split_on_pocket_edge(self) -> None:
        """
        When an polygon island touches the exterior at a single point the vorinoi
        diagram will have a vertex at that point. This leads code like _combine_edges
        to presume the edge passes through this vertex.
        It is unlikely this is the desired behavior so this method "shrinks" the
        voronoi edges slightly, creating 2 very close vertexes which stop the
        voronoi edges from touching.
        """
        split_at = []
        for vertex, edges_i in self.vertex_to_edges.items():
            point = Point(vertex)
            if len(edges_i) == 2 and self.distance_from_geom(point) <= EPS:
                split_at.append(vertex)

        while split_at:
            vertex = split_at.pop()
            edges_i = self.vertex_to_edges[vertex]

            assert len(edges_i) == 2

            edge_i_0 = edges_i[0]
            edge_0 = self.edges[edge_i_0]
            if edge_0.coords[0] == vertex:
                edge_0 = LineString(edge_0.coords[::-1])
            vertex_updated_0 = edge_0.interpolate(edge_0.length - EPS)

            edge_i_1 = edges_i[1]
            edge_1 = self.edges[edge_i_1]
            if edge_1.coords[0] == vertex:
                edge_1 = LineString(edge_1.coords[::-1])
            vertex_updated_1 = edge_1.interpolate(edge_1.length - EPS)

            self._remove_edge(edge_i_0)
            self._remove_edge(edge_i_1)

            self._store_edge(LineString([edge_0.coords[0], vertex_updated_0]), edge_i_0)
            self._store_edge(LineString([edge_1.coords[0], vertex_updated_1]), edge_i_1)

    def _vertexes_to_edge(self, vert_a: Vertex, vert_b: Vertex) -> Optional[int]:
        """
        Args:
            vert_a: First vertex.
            vert_b: Second vertex.
        Returns:
            The edge between vert_a and vert_b
            or None if no such edge exists.
        """
        edge = set(self.vertex_to_edges[vert_a]).intersection(set(self.vertex_to_edges[vert_b]))
        if edge:
            return edge.pop()
        return None

    def _get_cleanup_candidates(self, edge_i: int) -> Set[int]:
        """ Given an edge as a starting point, follow connected voronoi edges if they are likely candidates to delete.
        Args:
            edge_i: An edge index to start at. One end of the starting edge should have no other edge attached.
        Returns:
            A set of edge indexes that are candidates for pruning.
            In the event that an invalid starting position was supplied for edge_i, an empty set will be returned.
        """
        vertices_to_cleanup = []
        vertex_origin = None
        visited_edges = set()
        last_edge_angle = None
        next_edges = set([edge_i])
        previous_vertex_end = None

        while True:
            best_fit_edge_i = None
            best_fit_edge_angle = None
            best_fit_angle_diff = 0.3

            # Iterate through all edges adjoining the current candidate's candidate_vertex_end.
            for candidate_edge_i in next_edges:
                candidate_vertex_start, candidate_vertex_end = self.edge_to_vertex[candidate_edge_i]

                if last_edge_angle is None:
                    # First time around the loop.
                    if len(self.vertex_to_edges[candidate_vertex_start]) > 1:
                        # Try to start at the end of the edge with no junctions.
                        candidate_vertex_start, candidate_vertex_end = (
                                candidate_vertex_end, candidate_vertex_start)
                    if len(self.vertex_to_edges[candidate_vertex_start]) > 1:
                        # Trying to start from an edge with junctions at both ends.
                        # This is not a valid starting point.
                        return set()

                    vertex_origin = candidate_vertex_start
                    previous_vertex_end = candidate_vertex_start
                    vertices_to_cleanup.append(candidate_vertex_start)
                else:
                    # Make sure ends are correctly orientated.
                    if candidate_vertex_end == previous_vertex_end:
                        candidate_vertex_start, candidate_vertex_end = candidate_vertex_end, candidate_vertex_start
                assert candidate_vertex_start == previous_vertex_end
                assert vertices_to_cleanup[-1] == candidate_vertex_start

                if (abs(LineString([vertex_origin, candidate_vertex_end]).length) >=
                        abs(self.distance_from_geom(Point(candidate_vertex_end)) * self.tolerence)):
                    # If the distance from the start of the original edge to the end of the current candidate
                    # is longer than the distance to the closest polygon perimeter edge,
                    # then the considered edge is a valid voronoi edge and should not be pruned.
                    continue

                candidate_edge_angle = math.atan2(
                    (candidate_vertex_start[0] - candidate_vertex_end[0]),
                    (candidate_vertex_start[1] - candidate_vertex_end[1]))

                if last_edge_angle is None:
                    # First time around the loop.
                    best_fit_edge_angle = candidate_edge_angle
                    best_fit_vertex_end = candidate_vertex_end
                    best_fit_edge_i = candidate_edge_i
                    continue

                # When considering the next candidate to prune, it should be roughly co-linear
                # to the current candidate.
                angle_diff = abs((candidate_edge_angle - last_edge_angle))
                if angle_diff < best_fit_angle_diff:
                    best_fit_angle_diff = angle_diff
                    best_fit_vertex_end = candidate_vertex_end
                    best_fit_edge_i = candidate_edge_i
                    best_fit_edge_angle = candidate_edge_angle

            if best_fit_edge_i is None:
                break

            last_edge_angle = best_fit_edge_angle
            previous_vertex_end = best_fit_vertex_end
            vertices_to_cleanup.append(previous_vertex_end)
            visited_edges.add(best_fit_edge_i)

            next_edges = set(self.vertex_to_edges[previous_vertex_end]).difference(visited_edges)


        # Convert vertexes to edges.
        last_vertex = None
        to_cleanup: Set[int] = set()
        for vertex in vertices_to_cleanup:
            if last_vertex:
                assert self._vertexes_to_edge(last_vertex, vertex) is not None
                to_cleanup.add(self._vertexes_to_edge(last_vertex, vertex))
            last_vertex = vertex

        return to_cleanup

    def _drop_irrelevant_edges(self, preserve: Set[Tuple[float, float]]) -> None:
        """
        If any geometry resulting from a voronoi edge will be covered by the
        geometry of some other voronoi edge it is deemed irrelevant and pruned
        by this function.
        Args:
            preserve: A set of vertexes to be left somewhere in the diagram despite pruning.
        """
        to_prune: Set[int] = set()
        for edge_i in self.edges:
            to_prune |= self._get_cleanup_candidates(edge_i)

        # Any edges with a vertex in the preserve set should be considered more carefully.
        for preserve_vertex in preserve:
            to_preserve = set()
            still_prune = set()
            for edge_i in to_prune:
                vert_a, vert_b = self.edge_to_vertex[edge_i]
                if preserve_vertex in (vert_a, vert_b):
                    if len(self.vertex_to_edges[vert_a]) > 1 and len(self.vertex_to_edges[vert_b]) > 1:
                        # The edge being considered is connected to the main voronoi diagram at both ends.
                        # It should not be pruned.
                        to_preserve.add(edge_i)
                    else:
                        still_prune.add(edge_i)
            for edge_i in to_preserve:
                to_prune.remove(edge_i)
            if not to_preserve and still_prune:
                # Looks like we don't intend to preserve any of the requested vertices.
                # Pick one at random to preserve.
                to_prune.remove(still_prune.pop())

        # The self._get_cleanup_candidates algorithm does not check it isn't suggesting
        # the deletion of an edge that connects some other edge to the main diagram.
        # This could lead to orphaned islands of edges.
        # Here we make sure we only actually remove edges that do not link other
        # edges to the main diagram.
        while True:
            prune_edge = None
            for edge_i in to_prune:
                vert_a, vert_b = self.edge_to_vertex[edge_i]
                if len(self.vertex_to_edges[vert_a]) == 1:
                    prune_edge = edge_i
                    break
                if len(self.vertex_to_edges[vert_b]) == 1:
                    prune_edge = edge_i
                    break
            if prune_edge is None:
                break
            to_prune.remove(prune_edge)
            self._remove_edge(prune_edge)

    def widest_gap(self) -> Tuple[Point, float]:
        """
        Find the point inside self.polygon but furthest from an edge.

        Returns:
            (Point, float): Point at center of widest point.
                            Radius of circle that fits in widest point.
        """
        max_dist = max(abs(self.polygon.bounds[2] - self.polygon.bounds[0]),
                       abs(self.polygon.bounds[3] - self.polygon.bounds[1])) + 1

        widest_dist = 0
        widest_point = None
        for vertex in self.vertex_to_edges:
            nearest_dist = max_dist
            for ring in [self.polygon.exterior] + list(self.polygon.interiors):
                dist = Point(vertex).distance(ring)
                if dist < nearest_dist:
                    nearest_dist = dist
            if nearest_dist > widest_dist:
                widest_dist = nearest_dist
                widest_point = Point(vertex)
        return (widest_point, widest_dist)

    def vertex_on_perimiter(self) -> Point:
        """
        Find a point as close to the perimeter as possible.
        """
        bounding_box = box(*self.polygon.bounds).exterior
        closest = None
        distance = self.max_dist
        for vertex in self.vertex_to_edges:
            if bounding_box.intersects(Point(vertex)):
                return Point(vertex)
            dist = bounding_box.distance(Point(vertex))
            if dist < distance:
                distance = dist
                closest = Point(vertex)
        return closest

    def set_starting_point(self, point: Point) -> None:
        """
        Set a starting point within the polygon from which to start iterating over
        the voronoi diagram.
        If this point is not on the existing voronoi diagram it will need a new edge
        to be added to the diagram to connect this starting point.
        """
        if not point.within(self.polygon):
            raise PointOutsidePart(f"{point} not inside geometry")

        if point.coords[0] in self.vertex_to_edges:
            # point already exists on voronoi diagram. Nothing to do.
            self.start_point = point
            self.start_distance = self.distance_from_geom(point)
            return

        # Find closest voronoi edge to point but only across self.polygon.
        # Closest distances that go outside self.polygon are not valid.
        closest_edge = None
        closest_edge_index = None
        closest_dist = None
        new_vertex = None
        new_edge = None
        for edge_index, edge in self.edges.items():
            dist = edge.distance(point)
            proposed_new_vertex = round_coord(nearest_points(edge, point)[0].coords[0])
            proposed_new_edge = LineString([point, proposed_new_vertex])
            if closest_dist is None or dist < closest_dist:
                if proposed_new_edge.within(self.polygon) or proposed_new_edge.length == 0:
                    closest_dist = dist
                    closest_edge_index = edge_index
                    closest_edge = edge
                    new_vertex = proposed_new_vertex
                    new_edge = proposed_new_edge

        assert closest_edge_index is not None

        # Add the new vertex on the voronoi graph.
        v1, v2 = self.edge_to_vertex[closest_edge_index]
        new_edge_1 = LineString([v1, new_vertex])
        new_edge_2 = LineString([v2, new_vertex])
        self._remove_edge(closest_edge_index)
        if new_edge_1.length > 0:
            self._store_edge(new_edge_1)
        if new_edge_2.length > 0:
            self._store_edge(new_edge_2)

        # New vertex
        if new_edge is None or new_edge.length == 0:
            assert point.equals(Point(new_vertex))
            self.start_point = Point(new_vertex)
            self.start_distance = self.distance_from_geom(Point(new_vertex))
            return

        # Add a new edge to the voronoi graph.
        # This will home the new starting point.
        assert not point.equals(Point(new_vertex))
        self._store_edge(new_edge)
        self.start_point = point
        self.start_distance = self.distance_from_geom(point)
        return


def start_point_widest(
        desired_radius: Optional[float],
        step: float,
        pocket: BaseGeometry,
        already_cut: Optional[Polygon] = None
        ) -> VoronoiCenters:
    """
    Calculate voronoi diagram with the start point to be in the widest space,
    furthest from any part edge.
    Used when performing pocketing machining operations when cuts should start
    in the area with the most material to be removed. IE: furthest from the part
    perimeter.

    Args:
        desired_radius: Size of desired starting hole.
        step: Pocket algorithm's step size. already_cut will be oversizedby this amount.
        pocket: The area the returned voronoi diagram will be calculated for.
        already_cut: The area which must contain the starting hole.

    Returns:
        An instance of VoronoiCenters.
        Of particular interest are the .start_point and .max_desired_radius parameters.
    """
    if already_cut is None:
        complete_pocket = pocket
    else:
        complete_pocket = already_cut.union(pocket)

    voronoi = VoronoiCenters(complete_pocket, preserve_widest=True)
    start = voronoi.start_point
    if desired_radius:
        start = start.buffer(desired_radius)
    if already_cut and not start.within(already_cut.buffer(-step / 2)):
        # Start point outside cut area.
        # Generate a voronoi diagram of just the cut area and take the
        # start point from that.
        cut_voronoi = VoronoiCenters(already_cut, preserve_widest=True)
        voronoi = VoronoiCenters(complete_pocket, starting_point=cut_voronoi.start_point)
        del cut_voronoi

    return voronoi


def start_point_perimeter(
        desired_radius: float,
        pocket: BaseGeometry,
        already_cut: Polygon
        ) -> VoronoiCenters:
    """
    Calculate voronoi diagram with the start point to be inside the cut area,
    adjacent to the perimeter.
    Used when performing outer-peel machining operations when cuts start at the
    outside and work in towards the perimeter of the part.

    Args:
        desired_radius: Size of desired starting hole.
        pocket: The area the returned voronoi diagram will be calculated for.
        already_cut: The area which must contain the starting hole.

    Returns:
        An instance of VoronoiCenters.
        Of particular interest are the .start_point and .max_desired_radius parameters.
    """
    complete_pocket = already_cut.union(pocket)

    voronoi = VoronoiCenters(complete_pocket, preserve_edge=True)

    perimiter_point = voronoi.start_point
    assert perimiter_point is not None
    voronoi_edge_index = voronoi.vertex_to_edges[perimiter_point.coords[0]]
    assert len(voronoi_edge_index) == 1
    voronoi_edge = voronoi.edges[voronoi_edge_index[0]]
    cut_edge_section = already_cut.intersection(voronoi_edge)
    if cut_edge_section.length > desired_radius * 2:
        new_start_point = cut_edge_section.interpolate(0.5, normalized=True)

        voronoi = VoronoiCenters(complete_pocket, starting_point=new_start_point)
    else:
        # Can't fit desired_radius here.
        # Look for widest point in already_cut area.
        cut_voronoi = VoronoiCenters(already_cut, preserve_widest=True)
        voronoi = VoronoiCenters(complete_pocket, starting_point=cut_voronoi.start_point)
        del cut_voronoi

    return voronoi




"""
    A CAM library for generating HSM "peeling" toolpaths from supplied geometry.

    Copyright (C) <2022>  <duncan law>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

    mrdunk@gmail.com
"""

# pylint: disable=attribute-defined-outside-init

from typing import Dict, Generator, List, NamedTuple, Optional, Set, Tuple, Union

import copy
from enum import Enum
import math
import time

from shapely.affinity import rotate  # type: ignore
from shapely.geometry import box, GeometryCollection, LinearRing, LineString, MultiLineString, MultiPoint, MultiPolygon, Point, Polygon  # type: ignore
from shapely.ops import linemerge, split, unary_union  # type: ignore
from shapely.errors import GeometryTypeError

from hsm_nibble.debug import Display
from hsm_nibble.voronoi_centers import (
        round_coord, start_point_perimeter, start_point_widest, VoronoiCenters)  # type: ignore
from hsm_nibble.helpers import log  # type: ignore
DEBUG_DISPLAY = Display()


# Filter arcs that are entirely within this distance of a pocket edge.
SKIP_EDGE_ARCS = 1 / 20

# Trivially short lines joining arcs should be filtered.
SHORTEST_RAPID = 1e-6

# Number of tries before we give up trying to find a best-fit arc and just go
# with the best we have found so far.
ITERATION_COUNT = 50

# Whether to visit short voronoi edges first (True) or try to execute longer
# branches first.
# TODO: We could use more long range path planning that take the shortest total
# path into account.
#BREADTH_FIRST = True
BREADTH_FIRST = False

# When arc sizes drop below a certain point, we need to reduce the step size or
# forward motion due to the distance between arcs (step) becomes more than the
# arc diameter.
# This constant is the minimum arc radius, expressed as a multiple of the overlap
# size, size at which we start reducing step size.
CORNER_ZOOM = 2.0
# This constant is how much effect the feature will have. A value of "1" will
# keep the distance between each arc center proportional to the arc size.
CORNER_ZOOM_EFFECT = 1.0


class ArcDir(Enum):
    CW = 0
    CCW = 1
    Closest = 2

class MoveStyle(Enum):
    RAPID_OUTSIDE = 0
    RAPID_INSIDE = 1
    CUT = 2

class StartPointTactic(Enum):
    PERIMETER = 0  # Starting point hint for outer peel. On the outer perimeter.
    WIDEST = 1     # Starting point hint for inner pockets.


ArcData = NamedTuple("Arc", [
    ("origin", Point),
    ("radius", Optional[float]),
    ("start", Optional[Point]),
    ("end", Optional[Point]),
    ("start_angle", Optional[float]),
    ("span_angle", Optional[float]),
    ("winding_dir", Optional[ArcDir]),
    # TODO: ("widest_at", Optional[Point]),
    # TODO: ("start_DOC", float),
    # TODO: ("end_DOC", float),
    # TODO: ("widest_DOC", float),
    ("path", LineString),
    ("debug", Optional[str])
])

LineData = NamedTuple("Line", [
    ("start", Point),
    ("end", Point),
    ("path", LineString),
    ("move_style", MoveStyle),
])


def clean_linear_ring(ring: LinearRing) -> LinearRing:
    """ Remove duplicate points in a LinearRing. """
    new_ring = []
    prev_point = None
    first_point = None
    for point in ring.coords:
        if first_point is None:
            first_point = point
        if point == prev_point:
            continue
        else:
            new_ring.append(point)
            prev_point = point
    assert prev_point == first_point  # This is a loop.

    return LinearRing(new_ring)

def clean_polygon(polygon: Polygon) -> Polygon:
    exterior = clean_linear_ring(polygon.exterior)
    holes = []
    for hole in polygon.interiors:
        holes.append(clean_linear_ring(hole))

    return Polygon(exterior, holes=holes)

def clean_multipolygon(multi: MultiPolygon) -> MultiPolygon:
    if multi.geom_type != "MultiPolygon":
        multi = MultiPolygon([multi])
    polygons = []
    for polygon in multi.geoms:
        polygons.append(clean_polygon(polygon))

    return MultiPolygon(polygons)

def create_circle(origin: Point, radius: float, winding_dir: Optional[ArcDir] = None) -> ArcData:
    """
    Generate a circle that will be split into arcs to be part of the toolpath later.
    """
    span_angle = 2 * math.pi
    return ArcData(
        origin, radius, None, None, 0, span_angle, winding_dir, origin.buffer(radius).boundary, "")

def create_arc(
        origin: Point,
        radius: float,
        start_angle: float,
        span_angle: float,
        winding_dir: ArcDir) -> Optional[ArcData]:
    """
    Generate a arc.

    Args:
        origin: Center of arc.
        radius: Radius of arc.
        start_angle: Angle from vertical. (Clockwise)
        span_angle: Angular length of arc.
    """
    if radius == 0:
        return None

    span_angle = min(span_angle, 2 * math.pi)
    span_angle = max(span_angle, -2 * math.pi)

    start_angle = start_angle % (2 * math.pi)

    line_up = LineString([origin, Point(origin.x, origin.y + radius * 2)])
    circle_path = origin.buffer(radius).boundary
    circle_path = split(circle_path, line_up)
    points = circle_path.geoms[1].coords[:] + circle_path.geoms[0].coords[:]
    circle_path = LineString(points)

    if abs(span_angle) == 2 * math.pi:
        return ArcData(
                origin,
                radius,
                Point(circle_path.coords[0]),
                Point(circle_path.coords[0]),
                start_angle,
                span_angle,
                winding_dir,
                circle_path,
                "")

    # With shapely.affinity.rotate(...) -ive angles are clockwise.
    right_border = rotate(line_up, -span_angle, origin=origin, use_radians=True)

    if winding_dir == ArcDir.CCW:
        arc_path = split(circle_path, right_border).geoms[1]
        arc_path = LineString(arc_path.coords[::-1])
    else:
        arc_path = split(circle_path, right_border).geoms[0]

    arc_path = rotate(arc_path, -start_angle, origin=origin, use_radians=True)
    return ArcData(
            origin,
            radius,
            Point(arc_path.coords[0]),
            Point(arc_path.coords[-1]),
            start_angle,
            span_angle,
            winding_dir,
            arc_path,
            "")

def create_arc_from_path(
        origin: Point,
        path: LineString,
        radius: float,
        winding_dir: Optional[ArcDir] = None,
        debug: str = None
) -> ArcData:
    """
    Save data for the arc sections of the path.
    """
    start = Point(path.coords[0])
    end = Point(path.coords[-1])
    start_angle = None
    span_angle = None

    return ArcData(
            origin,
            radius,
            start,
            end,
            start_angle,
            span_angle,
            winding_dir,
            path,
            debug)

def mirror_arc(
        x_line: float,
        arc_data: ArcData,
        winding_dir: Optional[ArcDir] = None
        ) -> ArcData:
    """
    Mirror X axis of an arc about the origin point.
    """

    if winding_dir is None:
        assert (arc_data.winding_dir == ArcDir.CW and
                arc_data.span_angle is not None and 
                arc_data.span_angle > 0 or
                arc_data.winding_dir == ArcDir.CCW and
                arc_data.span_angle is not None and 
                arc_data.span_angle < 0)

        if arc_data.winding_dir == ArcDir.CW:
            winding_dir = ArcDir.CCW
        else:
            winding_dir = ArcDir.CW
    else:
        if winding_dir == arc_data.winding_dir:
            return arc_data

    arc_path = LineString(
            [((2 * x_line - point[0]), point[1]) for point in arc_data.path.coords]
            )
    origin = Point(2 * x_line - arc_data.origin.x, arc_data.origin.y)
    
    return ArcData(
            origin,
            arc_data.radius,
            Point(arc_path.coords[0]),
            Point(arc_path.coords[-1]),
            -arc_data.start_angle % (math.pi * 2) if arc_data.start_angle is not None else None,
            -arc_data.span_angle if arc_data.span_angle is not None else None,
            winding_dir,
            arc_path,
            arc_data.debug)

def complete_arc(
        arc_data: ArcData,
        winding_dir_: ArcDir = None
        ) -> Optional[ArcData]:
    """
    Calculate start_angle, span_angle and radius.
    Fix start, end and path direction based on winding_dir.
    This is called a lot so any optimizations here save us time.
    Given some properties of an arc, calculate the others.
    """
    winding_dir = winding_dir_
    if winding_dir is None:
        winding_dir = arc_data.winding_dir
    if winding_dir is None or winding_dir == ArcDir.Closest:
        winding_dir = ArcDir.CW

    # Make copy of path since we may need to modify it.
    path = LineString(arc_data.path)
    if path.length == 0.0:
        return None

    start_coord = path.coords[0]
    end_coord = path.coords[-1]
    mid = path.interpolate(0.5, normalized=True)

    # Breaking these out once rather than separately inline later saves us ~7%
    # CPU time overall.
    org_x, org_y = arc_data.origin.xy
    start_x, start_y = start_coord
    mid_x, mid_y = mid.xy
    end_x, end_y = end_coord

    start_angle = math.atan2(start_x - org_x[0], start_y - org_y[0])
    end_angle = math.atan2(end_x - org_x[0], end_y - org_y[0])
    mid_angle = math.atan2(mid_x[0] - org_x[0], mid_y[0] - org_y[0])

    ds = (start_angle - mid_angle) % (2 * math.pi)
    de = (mid_angle - end_angle) % (2 * math.pi)
    if ((ds > 0 and de > 0 and winding_dir == ArcDir.CCW) or
            (ds < 0 and de < 0 and winding_dir == ArcDir.CW)):
        # Needs reversed.
        path = LineString(path.coords[::-1])
        start_angle, end_angle = end_angle, start_angle
        start_coord, end_coord = end_coord, start_coord

    if winding_dir == ArcDir.CW:
        span_angle = (end_angle - start_angle) % (2 * math.pi)
    elif winding_dir == ArcDir.CCW:
        span_angle = (-(start_angle - end_angle) % (2 * math.pi)) - (2 * math.pi)

    if span_angle == 0.0:
        span_angle = 2 * math.pi

    radius = arc_data.radius or arc_data.origin.distance(Point(path.coords[0]))

    return ArcData(
            arc_data.origin,
            radius,
            Point(start_coord),
            Point(end_coord),
            start_angle % (2 * math.pi),
            span_angle,
            winding_dir,
            path,
            arc_data.debug)


def arcs_from_circle_diff(
        circle: ArcData,
        already_cut: Polygon,
        debug: str = None
        ) -> List[ArcData]:
    """ Return any sections of circle that do not overlap already_cut. """
    if not already_cut:
        return [circle]
    if circle is None:
        return None

    line_diff = circle.path.difference(already_cut)
    if not line_diff:
        return []
    if line_diff.geom_type == "MultiLineString":
        line_diff = linemerge(line_diff)
    if line_diff.geom_type != "MultiLineString":
        line_diff = MultiLineString([line_diff])

    arcs = []
    assert circle.radius is not None
    for arc in line_diff.geoms:
        arcs.append(create_arc_from_path(
            circle.origin, arc, circle.radius, winding_dir=circle.winding_dir, debug=debug))
    return arcs


def _colapse_dupe_points(line: LineString) -> Optional[LineString]:
    """
    Filter out duplicate points.
    TODO: Profile whether a .simplify(0) would be quicker?
    """
    points = []
    last_point = None
    for point in line.coords:
        if last_point == point:
            continue
        points.append(point)
        last_point = point
    if len(points) < 2:
        return None
    return LineString(points)


def split_line_by_poly(line: LineString, poly: Union[Polygon, MultiPolygon]) -> MultiLineString:
    """
    split() sometimes fails if line shares a point with one of poly's rings.
    """

    if isinstance(poly, Polygon):
        poly = MultiPolygon([poly])
    rings = []
    for p in poly.geoms:
        rings.append(p.exterior)
        rings.extend(p.interiors)
    new_lines = []
    lines = [line]
    for ring in rings:
        for l in lines:
            try:
                split_line = split(l, LineString(ring))
            except (GeometryTypeError, TypeError, ValueError):
                split_line = MultiLineString([l])
            new_lines.extend(split_line.geoms)
        lines = new_lines
        new_lines = []

    # Check for gaps caused by line being co-linear to poly.
    last_new_line = None
    checked_lines = []
    for new_line in lines:
        if last_new_line is None:
            if line.coords[0] != new_line.coords[0]:
                checked_lines.append(LineString([line.coords[0], new_line.coords[0]]))
        else:
            if last_new_line.coords[-1] != new_line.coords[0]:
                checked_lines.append(LineString([last_new_line.coords[-1], new_line.coords[0]]))
        checked_lines.append(new_line)
        last_new_line = new_line
    if last_new_line.coords[-1] != line.coords[-1]:
        checked_lines.append(LineString([last_new_line.coords[-1], line.coords[-1]]))

    return MultiLineString(checked_lines)


class BaseGeometry:
    # Area we have calculated arcs for. Calculated by appending full circles.
    calculated_area_total: Polygon
    # Area we have stored the toolpath for. Calculated by appending full circles.
    cut_area_total: Polygon

    starting_cut_area: Polygon

    def __init__(
            self,
            to_cut: Union[Polygon, MultiPolygon],
            step: float,
            winding_dir: ArcDir,
            already_cut: Polygon = None,
    ) -> None:
        self.polygon: Union[Polygon, MultiPolygon] = to_cut
        self.step: float = step
        self.winding_dir: ArcDir = winding_dir
        if already_cut is None:
            already_cut = Polygon()
        self.starting_cut_area = already_cut
        self.path: List[Union[ArcData, LineData]] = []
        self.pending_arc_queues: List[List[ArcData]] = []
        self.last_arc: Optional[ArcData] = None

        self.calculated_area_total = self.starting_cut_area
        self.cut_area_total = copy.deepcopy(self.starting_cut_area)
        assert self.calculated_area_total is not self.cut_area_total

        self._filter_input_geometry()
        self.dilated_starting_polygon = self.polygon.buffer(self.step / 10)

    def _filter_input_geometry(self) -> None:
        """Remove any tiny polygons from a MultiPolygon. """
        min_area = (self.step / 20) ** 2

        if self.polygon.geom_type != "MultiPolygon":
            self.polygon = MultiPolygon([self.polygon])

        self.polygon = MultiPolygon([poly for poly in self.polygon.geoms
                    if poly.buffer(-self.step / 20).area > min_area])

    def _flush_arc_queues(self) -> None:
        while self.pending_arc_queues:
            to_process = self.pending_arc_queues.pop(0)
            self._arcs_to_path(to_process)

    def _check_queue_overlap(self, arc: ArcData, queue_index: int) -> bool:
        """
        Arcs should only be added if the area inside them has been cleared.
        This checks for overlaps between the specified arc and those that are
        to be cut /after/ it.
        Returns:
            True: If no overlap occurs.
            False: If the proposed arc is blocked for this queue by as yet uncut arcs.
        """
        dilated_arc = arc.path.buffer(self.step)
        for queue in self.pending_arc_queues[queue_index + 1:]:
            for existing_arc in queue:
                if existing_arc.path.intersects(dilated_arc):
                    return False
        return True

    def _queue_arcs(self, new_arcs: List[ArcData]) -> None:
        """
        When an arc intersects with an area that has already been cut the arc may
        get split into multiple pieces.
        When we cone to join the arcs we want to join the "left" arcs to each other
        and the "right" arcs to each other. (There may be more than 2 sets as well.)
        To do this we need to store arcs in separate queues. Each queue contains
        arcs that should be joined to each other.
        """

        # Need to put each arc in the queue with nearest predecessor.
        modified_queues_indexes = set()

        if len(new_arcs) == 1 and len(self.pending_arc_queues) == 1:
            # Optimization to save expensive repeated distance calculations.
            # Only a single queue and a single arc so we don't need complicated
            # queues.
            # Just add the arc to the single queue.
            # Since there are no other queues depending on our remaining one,
            # it is safe to drain it,
            self.pending_arc_queues[0].append(new_arcs[0])
            to_process = self.pending_arc_queues.pop(0)
            self._arcs_to_path(to_process)
            return
        else:
            closest_queue: Optional[List[ArcData]]
            for arc in new_arcs:
                closest_queue = None
                closest_queue_index = None
                closest_dist = self.step
                for queue_index, queue in enumerate(self.pending_arc_queues):
                    if self.winding_dir == ArcDir.Closest:
                        combined_dist = 1
                        seperate_dist = 0
                    else:
                        assert arc.start is not None
                        assert queue[-1].start is not None
                        assert queue[-1].end is not None

                        seperate_dist = (
                                queue[-1].start.distance(queue[-1].end) + arc.start.distance(arc.end))
                        combined_dist = (
                                queue[-1].start.distance(arc.end) + queue[-1].end.distance(arc.start))

                    seperation = arc.path.distance(queue[-1].path)
                    if seperation < closest_dist or combined_dist < seperate_dist:
                        if self._check_queue_overlap(arc, queue_index):
                            closest_dist = seperation
                            closest_queue = queue
                            closest_queue_index = queue_index
                if closest_queue is None:
                    # Not close to any predecessor. Create new queue.
                    closest_queue = []
                    closest_queue_index = len(self.pending_arc_queues)
                    self.pending_arc_queues.append(closest_queue)
                closest_queue.append(arc)
                modified_queues_indexes.add(closest_queue_index)
                assert closest_queue_index is not None
                assert closest_queue is self.pending_arc_queues[closest_queue_index]
                assert arc in closest_queue

        # Queues need processed in the order they were created: FIFO.
        # It is only safe to process the oldest queue (index: 0) as any younger
        # queue may be a child of it.
        # TODO: If we really wanted to, whenever there are any un-modified queues,
        # we could process the contents of all the older queues even though they
        # are still being appended to before processing the un-modifies queue(s).
        if modified_queues_indexes and 0 not in modified_queues_indexes:
            to_process = self.pending_arc_queues.pop(0)
            self._arcs_to_path(to_process)

    def _arcs_to_path(self, arcs: List[ArcData]) -> None:
        """
        Process list list of arcs, calculate tool path to join one to the next
        and apply them to the self.path parameter.

        Note: This function modifies the arcs parameter in place.
        """
        while arcs:
            arc = arcs.pop(0)
            if arc is None:
                continue

            winding_dir = self.winding_dir
            last_arc = self.last_arc
            if winding_dir == ArcDir.Closest:
                if last_arc is None:
                    winding_dir = ArcDir.CW
                else:
                    # TODO: We could improve this: Rather that taking the opposite
                    # of the last arc, we could work out the closest end based on
                    # the last drawn arc.
                    if last_arc.winding_dir == ArcDir.CCW:
                        winding_dir = ArcDir.CW
                    else:
                        winding_dir = ArcDir.CCW

                    assert last_arc.start is not None
                    assert last_arc.end is not None

                    if (last_arc.end.distance(arc.start) <
                            last_arc.end.distance(arc.end)):
                        winding_dir = ArcDir.CW
                    if (last_arc.start.distance(arc.end) <
                            last_arc.end.distance(arc.end)):
                        winding_dir = ArcDir.CW

                # TODO: Just reversing arc.path, arc.span_angle and arc.winding_dir
                # would be quicker than calling complete_arc(...).
                arc_ = complete_arc(arc, winding_dir)
                if arc_ is None:
                    continue
                arc = arc_

            assert arc.path.length > 0
            assert len(arc.path.coords) >= 2
            assert arc.span_angle != 0

            if last_arc is not None:
                self.path += self.join_arcs(arc)
            self.path.append(arc)
            self.last_arc = arc

    def join_arcs(self, next_arc: ArcData) -> List[LineData]:
        """
        Generate CAM tool path to join the end of one arc to the beginning of the next.
        Also tracks what geometry has actually been cut (as opposed to just planned).
        """
        assert self.last_arc

        # Add a section of the full circle to the cut area.
        arc_poly = Polygon(
                list(next_arc.path.coords) + [(next_arc.origin.x, next_arc.origin.y)]
                ).buffer(self.step / 20)
        to_cut_area_total = self.cut_area_total.union(arc_poly)

        lines = []
        path = LineString([self.last_arc.end, next_arc.start])
        inside_pocket = (
                path.covered_by(self.dilated_starting_polygon)
                or path.covered_by(to_cut_area_total)
                )

        if inside_pocket:
            # Whole path is inside pocket.
            if path.length <= self.step:
                return [LineData(
                        Point(path.coords[0]),
                        Point(path.coords[-1]),
                        path,
                        MoveStyle.CUT,
                        )]
            
            split_for_last = path.interpolate(-self.step)
            last_line = LineData(
                    Point(split_for_last),
                    Point(next_arc.start),
                    LineString([split_for_last, next_arc.start]),
                    MoveStyle.CUT,
                    )

            remaining_path = LineString([self.last_arc.end, split_for_last])

            split_path = split_line_by_poly(remaining_path, self.cut_area_total)

            for part in split_path.geoms:
                assert part.geom_type == "LineString"
                assert len(part.coords) == 2

                move_style = MoveStyle.CUT
                if part.intersection(self.cut_area_total).length > part.length - self.step / 20:
                #if part.covered_by(self.cut_area_total):
                    move_style = MoveStyle.RAPID_INSIDE

                line = LineData(
                        Point(part.coords[0]),
                        Point(part.coords[-1]),
                        part,
                        move_style,)

                if line.path.length > SHORTEST_RAPID:
                    lines.append(line)
            lines.append(last_line)
        else:
            # Path is not entirely inside pocket or crosses uncut area.
            move_style = MoveStyle.RAPID_OUTSIDE
            line = LineData(self.last_arc.end, next_arc.start, path, move_style,)
            if line.path.length > SHORTEST_RAPID:
                lines.append(line)

        self.cut_area_total = to_cut_area_total

        return lines

    def _split_arcs(self, full_arcs: List[ArcData]) -> List[ArcData]:
        """
        When an arcs pass through an already cut area, it should be trimmed to
        remove the already cut section.
        It's important to keep the sequence of the order of the resulting sub arcs
        the same as the originals so the final path can pass through them in order.
        """
        calculated_area_total = self.calculated_area_total
        split_arcs = []
        for full_arc in full_arcs:
            new_arcs = arcs_from_circle_diff(full_arc, calculated_area_total)
            if not new_arcs:
                continue

            new_arcs = list(filter(
                None, [complete_arc(new_arc, new_arc.winding_dir)
                    for new_arc in new_arcs if new_arc.path.length]
                ))

            arc_set = set()
            for new_arc_index, new_arc in enumerate(new_arcs):
                arc_set.add((full_arc.path.project(new_arc.start), new_arc_index))

            for _, new_arc_index in sorted(arc_set):
                split_arcs.append(new_arcs[new_arc_index])

        return split_arcs


class Pocket(BaseGeometry):
    """
    A CAM library to generate a HSM "peeling" pocketing toolpath.
    """

    start_point: Point
    max_starting_radius: float
    starting_angle: Optional[float] = None
    last_arc: Optional[ArcData]
    last_circle: Optional[ArcData]
    debug: bool = False

    def __init__(
            self,
            to_cut: Polygon,
            step: float,
            winding_dir: ArcDir,
            already_cut: Polygon = None,
            generate: bool = False,
            voronoi: Optional[VoronoiCenters] = None,
            starting_point_tactic: StartPointTactic = StartPointTactic.WIDEST,
            starting_point: Point = None,
            starting_radius: float = None,
            debug: bool = False,
    ) -> None:
        """
        Arguments:
            to_cut: The area this pocket should cover.
            step: The distance between cutting path passes.
            winding_dir: Tactic for choosing clockwise or anticlockwise path sections.
            already_cut: An area where no cutting path should be generated but
                also no care needs to be taken when intersecting it. Also rapid moves
                may pass through it.
            generate: Whether to use the get_arcs(...) parameter as a generator
                function (True) or calculate the whole path at the start (False).
            voronoi: Optionally pass in a voronoi diagram detailing points equidistant
                from the part edges. If not provided, one will be created.
            starting_point_tactic: Tactic used to pick the very start of the cutting
                path.
            starting_point: Override the automatically calculated cutting path start point.
                Must be withing either the `to_cut` area or the `already_cut` area if
                one was specified.
            starting_radius: Optionally set the required space at the start_point.
                Used to make sure an entry helix has enough space at the start_point
                when machining.

        Properties:
            start_point: The start of the cutting path. If one is not specified with
                the starting_point parameter, this will be calculated automatically.
            max_starting_radius: The maximum radius a circle at start_point could
                be without overlapping the part's edges.
            starting_angle: `None` if the entry circle is completely within the already_cut
                area. Otherwise it contains the angle to the start of the cutting path
                where the entry circle intersects the cutting path.
        """
        if already_cut is None:
            already_cut = Polygon()
        complete_pocket = already_cut.union(to_cut)
        to_cut = complete_pocket.difference(already_cut)

        clean_multipolygon(to_cut)
        clean_multipolygon(complete_pocket)
        clean_multipolygon(already_cut)

        self.starting_cut_area = already_cut
        self.starting_radius = starting_radius
        self.generate = generate
        self.debug = debug

        # Calculate voronoi diagram for finding points equidistant between part edges.
        if voronoi is None:
            if starting_point is not None:
                voronoi = VoronoiCenters(complete_pocket, starting_point=starting_point)
            elif starting_point_tactic == StartPointTactic.WIDEST:
                voronoi = start_point_widest(
                        self.starting_radius, step, complete_pocket, already_cut)
            elif starting_point_tactic == StartPointTactic.PERIMETER:
                voronoi = start_point_perimeter(
                        self.starting_radius or step, complete_pocket, already_cut)

            assert voronoi is not None
        self.voronoi = voronoi

        # Calculate entry hole settings.
        self.max_starting_radius = voronoi.max_starting_radius
        if starting_radius and self.max_starting_radius < starting_radius:
            print(f"Warning: Starting radius of {starting_radius} overlaps boundaries "
                  f"at {voronoi.start_point}")
            print(f" Largest possible at this location: {round(self.max_starting_radius, 3)}")
        if self.starting_radius is not None:
            radius = min(self.starting_radius, self.max_starting_radius)
            start_circle = voronoi.start_point.buffer(radius)
            already_cut = already_cut.union(start_circle)

        super().__init__(to_cut, step, winding_dir, already_cut=already_cut)

        self._reset()
        self.calculate_path()

    def _reset(self) -> None:
        """ Cleanup and/or initialise everything. """

        self.start_point: Point = self.voronoi.start_point
        self.start_radius: float = self.voronoi.start_distance or 0

        self.arc_fail_count: int = 0
        self.path_fail_count: int = 0
        self.loop_count: int = 0

        self.visited_edges: Set[int] = set()
        self.open_paths: Dict[int, Tuple[float, float]] = {}

        self.path_len_progress: float = 0.0
        self.path_len_total: float = 0.0
        for edge in self.voronoi.edges.values():
            self.path_len_total += edge.length

        # Used to detect when an arc is too close to the edge to be worthwhile.
        self.dilated_polygon_boundaries = []
        multi = self.polygon
        if multi.geom_type != "MultiPolygon":
            multi = MultiPolygon([multi])
        for poly in multi.geoms:
            for ring in [poly.exterior] + list(poly.interiors):
                self.dilated_polygon_boundaries.append(
                        ring.buffer(self.step * SKIP_EDGE_ARCS))

        # Generate first circle from which other arcs expand.
        entry_circle = EntryCircle(
                self.polygon,
                self.start_point,
                self.start_radius,
                self.step,
                self.winding_dir,
                already_cut=self.calculated_area_total,
                path=self.path)
        entry_circle.spiral()
        entry_circle.circle()

        if self.starting_radius is not None:
            radius = min(self.starting_radius, self.max_starting_radius)
            if (self.path and
                    self.start_point.buffer(radius).distance(
                        self.path[0].start) < self.step / 20):
                assert self.path[0].start
                dx = self.path[0].start.x - self.start_point.x
                dy = self.path[0].start.y - self.start_point.y
                self.starting_angle = math.atan2(dx, dy)

        self.last_arc = entry_circle.last_arc

        self.last_circle: Optional[ArcData] = create_circle(
            self.start_point, self.start_radius)
        self.calculated_area_total = self.calculated_area_total.union(
                Polygon(self.last_circle.path))

    def calculate_path(self) -> None:
        """ Reset path and restart from beginning. """
        # Create the generator.
        generator = self.get_arcs()

        if not self.generate:
            # Don't want to use it as a generator so set it running.
            try:
                next(generator)
            except StopIteration:
                pass

    def done_generating(self):
        if DEBUG_DISPLAY:
            DEBUG_DISPLAY.display(
                    polygons={
                        "previously_cut": self.starting_cut_area,
                        "to_cut": self.polygon,
                        "to_cut_dilated": self.dilated_starting_polygon,
                        #"cut_progress": self.calculated_area_total,
                        },
                    voronoi=self.voronoi,
                    path=self.path
                    )

    def _choose_next_path(
            self,
            current_pos: Optional[Tuple[float, float]] = None
    ) -> Optional[Tuple[float, float]]:
        """
        Choose a vertex with an un-traveled voronoi edge leading from it.

        Returns:
            A vertex that has un-traveled edges leading from it.
        """
        # Cleanup.
        for edge_i in self.visited_edges:
            if edge_i in self.open_paths:
                self.open_paths.pop(edge_i)

        shortest = self.voronoi.max_dist + 1
        closest_vertex: Optional[Tuple[float, float]] = None
        closest_edge: Optional[int] = None
        for edge_i, vertex in self.open_paths.items():
            if current_pos:
                dist = Point(vertex).distance(Point(current_pos))
            else:
                dist = 0

            if closest_vertex is None:
                shortest = dist
                closest_vertex = vertex
                closest_edge = edge_i
                if not current_pos:
                    break
            elif dist < shortest:
                closest_vertex = vertex
                closest_edge = edge_i
                shortest = dist

        if closest_edge is not None:
            self.open_paths.pop(closest_edge)

        self.last_circle = None
        return closest_vertex

    @classmethod
    def _extrapolate_line(cls, extra: float, line: LineString) -> LineString:
        """
        Extend a line at both ends in the same direction it points.
        """
        coord_0, coord_1 = line.coords[:2]
        coord_m2, coord_m1 = line.coords[-2:]
        ratio_begin = extra / LineString([coord_0, coord_1]).length
        ratio_end = extra / LineString([coord_m2, coord_m1]).length
        coord_begin = Point(
            coord_0[0] + (coord_0[0] - coord_1[0]) * ratio_begin,
            coord_0[1] + (coord_0[1] - coord_1[1]) * ratio_begin)
        coord_end = Point(
            coord_m1[0] + (coord_m1[0] - coord_m2[0]) * ratio_end,
            coord_m1[1] + (coord_m1[1] - coord_m2[1]) * ratio_end)
        return LineString([coord_begin] + list(line.coords) + [coord_end])

    @classmethod
    def _converge(cls, kp: float) -> Generator[float, Tuple[float, float], None]:
        """
        Algorithm used for recursively estimating the position of the best fit arc.

        Arguments:
            kp: Proportional multiplier.
        Yields:
            Arguments:
                target: Target step size.
                current: step size resulting from the previous iteration result.
            next distance.
        Returns:
            Never exits Yield loop.
        """
        error: float = 0.0
        value: float = 0.0

        while True:
            target, current = yield value

            error = target - current
            prportional = kp * error
            value = prportional

    def _arc_at_distance(self, distance: float, voronoi_edge: LineString) -> Tuple[Point, float]:
        """
        Calculate the center point and radius of the largest arc that fits at a
        set distance along a voronoi edge.
        """
        pos = voronoi_edge.interpolate(distance)
        radius = self.voronoi.distance_from_geom(pos)

        return (pos, radius)

    def _furthest_spacing_arcs(self, arcs: List[ArcData], last_circle: ArcData) -> float:
        """
        Calculate maximum step_over between 2 arcs.
        """
        #return self._furthest_spacing_shapely(arcs, last_circle.path)
        spacing = -self.voronoi.max_dist

        for arc in arcs:
            spacing = max(spacing,
                          last_circle.origin.hausdorff_distance(arc.path) - last_circle.radius)

            #for index in range(0, len(arc.path.coords), 1):
            #    coord = arc.path.coords[index]
            #    spacing = max(spacing,
            #            Point(coord).distance(last_circle.origin) - last_circle.radius)

        return abs(spacing)

    @classmethod
    def _furthest_spacing_shapely(
            cls, arcs: List[ArcData], previous: LineString) -> float:
        """
        Calculate maximum step_over between 2 arcs.

        TODO: Current implementation is expensive. Not sure how shapely's distance
        method works but it is likely "O(N)", making this implementation N^2.
        We can likely reduce that to O(N*log(N)) with a binary search.

        Arguments:
            arcs: The new arcs.
            previous: The previous cut path geometry we are testing the arks against.

        Returns:
            The step distance.
        """
        spacing = -1
        polygon = previous
        for arc in arcs:
            if not arc.path:
                continue

            # This is expensive but yields good results.
            # Probably want to do a binary search version?

            for index in range(0, len(arc.path.coords), 1):
                coord = arc.path.coords[index]
                #spacing = max(spacing, Point(coord).distance(polygon))
                spacing = max(spacing, polygon.distance(Point(coord)))

        return spacing

    def _calculate_arc(
            self,
            voronoi_edge: LineString,
            start_distance: float,
            min_distance: float,
    ) -> Tuple[float, List[ArcData]]:
        """
        Calculate the arc that best fits within the path geometry.

        A given point on the voronoi_edge is equidistant between the edges of the
        desired cut path. We can calculate this distance and it forms the radius
        of an arc touching the cut path edges.
        We need the furthest point on that arc to be desired_step distance away from
        the previous arc. It is hard to calculate a point on the voronoi_edge that
        results in the correct spacing between the new and previous arc.

        The constraints for the new arc are:
        1) The arc must go through the point on the voronoi edge desired_step
          distance from the previous arc's intersection with the voronoi edge.
        2) The arc must be a tangent to the edge of the cut pocket.
          Or put another way: The distance from the center of the arc to the edge
          of the cut pocket should be the same as the distance from the center of
          the arc to the point described in 1).

        Rather than work out the new arc centre position with maths, it is quicker
        and easier to use a binary search, moving the proposed centre repeatedly
        and seeing if the arc fits.

        Arguments:
            voronoi_edge: The line of mid-way points between the edges of desired
              cut path.
            start_distance: The distance along voronoi_edge to start trying to
              find an arc that fits.
            min_distance: Do not return arcs below this distance; The algorithm
              is confused and traveling backwards.
        Returns:
            A tuple containing:
                1. Distance along voronoi edge of the final arc.
                2. A collection of ArcData objects containing relevant information
                about the arcs generated with an origin the specified distance
                allong the voronoi edge.
        """
        # A generator to converge on desired spacing.
        #converge = self._converge(0.75)
        converge = self._converge(0.76)
        converge.send(None)  # type: ignore

        coverage_algos = [
                #self._converge(0.75),
                self._converge(0.76),
                self._converge(0.74),
                self._converge(0.78),
                self._converge(0.72),
                self._converge(0.7),
                self._converge(0.8),
                ]

        color_overide = None

        desired_step = min(self.step, (voronoi_edge.length - start_distance))

        distance = start_distance + desired_step

        count: int = 0
        circle: Optional[ArcData] = None
        arcs: List[ArcData] = []
        progress: float = 0.0
        best_progress: float = 0.0
        best_distance: float = 0.0
        dist_offset: int = 100000
        corner_zoom = CORNER_ZOOM * self.step

        # Extrapolate line beyond it's actual distance to give the algorithm
        # room to overshoot while converging on an optimal position for the new arc.
        edge_extended: LineString = self._extrapolate_line(
            dist_offset, voronoi_edge)
        assert abs(edge_extended.length -
                   (voronoi_edge.length + 2 * dist_offset)) < 0.0001

        assert self.calculated_area_total
        assert self.calculated_area_total.is_valid

        # Loop multiple times, trying to converge on a distance along the voronoi
        # edge that provides the correct step size.
        while count <= ITERATION_COUNT:
            count += 1

            # Propose an arc.
            pos, radius = self._arc_at_distance(
                distance + dist_offset, edge_extended)
            circle = create_circle(pos, radius, self.winding_dir)

            # Compare proposed arc to cut area.
            # We are only interested in sections that have not been cut yet.
            arcs = self._split_arcs([circle])
            if not arcs:
                # arc is entirely hidden by previous cut geometry.

                if best_progress > 0:
                    # Has made some progress.
                    count = ITERATION_COUNT
                    color_overide = "orange"
                    break

                # Has not found any useful arc yet.
                # Don't record it as an arc that needs drawn.
                self.last_circle = circle
                return(distance, [])

            # Progress is measured as the furthest point the proposed arc is
            # from the previous one. We are aiming for proposed == desired_step.
            if self.last_circle:
                progress = self._furthest_spacing_arcs(arcs, self.last_circle)
            else:
                progress = self._furthest_spacing_shapely(arcs, self.calculated_area_total)

            desired_step = min(self.step, (voronoi_edge.length - start_distance))
            if radius < corner_zoom:
                # Limit step size as the arc radius gets very small.
                multiplier = (corner_zoom - radius) / corner_zoom
                desired_step = self.step * (1 - CORNER_ZOOM_EFFECT * multiplier)

            if abs(desired_step - progress) < abs(desired_step - best_progress):
                # Better fit.
                best_progress = progress
                best_distance = distance

                if abs(desired_step - progress) < desired_step / 20:
                    # Good enough fit.
                    best_progress = progress
                    best_distance = distance
                    break

            modifier = converge.send((desired_step, progress))
            distance += modifier

        if count == ITERATION_COUNT:
            color_overide = "red"
            if distance < min_distance:
                # Moving the wrong way along the voronoi edge.
                # Only happens when we've been to the end of an edge already.
                return (voronoi_edge.length, [])

        if best_distance > voronoi_edge.length:
            best_distance = voronoi_edge.length

        if distance != best_distance or progress != best_progress or color_overide is not None:
            distance = best_distance
            progress = best_progress
            pos, radius = self._arc_at_distance(
                distance + dist_offset, edge_extended)
            circle = create_circle(Point(pos), radius, self.winding_dir)
            arcs = self._split_arcs([circle])

        if count == ITERATION_COUNT and self.debug:
            # Log some debug data.
            distance_remain = voronoi_edge.length - distance
            self.arc_fail_count += 1
            log("\tDid not find an arc that fits. Spacing/Desired: "
                f"{round(progress, 3)}/{desired_step}"
                "\tdistance remaining: "
                f"{round(distance_remain, 3)}")

        self.loop_count += count

        assert circle is not None
        self.last_circle = circle
        self.calculated_area_total = self.calculated_area_total.union(
                Polygon(circle.path))

        filtered_arcs = []
        for arc in arcs:
            if self._filter_arc(arc):
                filtered_arcs.append(arc)

        return (distance, filtered_arcs)

    def _join_voronoi_branches(self, start_vertex: Tuple[float, float]) -> LineString:
        """
        Walk a section of the voronoi edge tree, creating a combined edge as we
        go.

        Returns:
            A LineString object of the combined edges.
        """
        vertex = start_vertex

        line_coords: List[Tuple[float, float]] = []

        while True:
            branches = self.voronoi.vertex_to_edges[vertex]
            candidate = None
            longest = 0
            for branch in branches:
                if branch not in self.visited_edges:
                    self.open_paths[branch] = vertex
                    length = self.voronoi.edges[branch].length
                    if candidate is None:
                        candidate = branch
                    elif BREADTH_FIRST and length < longest:
                        candidate = branch
                    elif not BREADTH_FIRST and length > longest:
                        candidate = branch

                    longest = max(longest, length)

            if candidate is None:
                break

            self.visited_edges.add(candidate)
            edge_coords = self.voronoi.edges[candidate].coords

            if not line_coords:
                line_coords = edge_coords
                if start_vertex != line_coords[0]:
                    line_coords = line_coords[::-1]
            else:
                if line_coords[-1] == edge_coords[-1]:
                    edge_coords = edge_coords[::-1]
                assert line_coords[0] == start_vertex
                assert line_coords[-1] == edge_coords[0]
                line_coords = list(line_coords) + list(edge_coords)

            vertex = line_coords[-1]

        line = LineString(line_coords)

        return _colapse_dupe_points(line)

    def get_arcs(self, timeslice: int = 0):
        """
        A generator method to create the path.

        Class instance properties:
            self.generate: bool: Whether or not to yield.
                False: Do not yield. Generate all data in one shot.
                True: Yield an estimated ratio of path completion.

        Arguments:
            timeslice: int: How long to generate arcs for before yielding (ms).
        """
        start_time = round(time.time() * 1000)  # ms

        start_vertex: Optional[Tuple[float, float]
                               ] = self.start_point.coords[0]
        while start_vertex is not None:
            # This outer loop iterates through the voronoi vertexes, looking for
            # a voronoi edge that has not yet had arcs calculated for it.
            combined_edge = self._join_voronoi_branches(start_vertex)
            if not combined_edge:
                start_vertex = self._choose_next_path()
                continue

            dist = 0.0
            best_dist = dist
            stuck_count = int(combined_edge.length * 10 / self.step + 10)
            while abs(dist - combined_edge.length) > self.step / 20 and stuck_count > 0:
                # This inner loop travels along a voronoi edge, trying to fit arcs
                # that are the correct distance apart.
                stuck_count -= 1
                dist, new_arcs = self._calculate_arc(combined_edge, dist, best_dist)

                self.path_len_progress -= best_dist
                self.path_len_progress += dist

                if dist < best_dist and False:
                    # Getting worse not better or staying the same.
                    # This can happen legitimately but is an indication the algorthm
                    # may be stuck.
                    stuck_count = int(stuck_count / 2)
                else:
                    best_dist = dist
                self._queue_arcs(new_arcs)

                if timeslice >= 0 and self.generate:
                    now = round(time.time() * 1000)  # (ms)
                    if start_time + timeslice < now:
                        yield min(0.999, self.path_len_progress / self.path_len_total)
                        start_time = round(time.time() * 1000)  # (ms)

            if stuck_count <= 0:
                print(
                    f"stuck: {round(dist, 2)} / {round(combined_edge.length, 2)}")
                self.path_fail_count += 1

            start_vertex = self._choose_next_path(combined_edge.coords[-1])

            self._flush_arc_queues()

        if timeslice and self.generate:
            yield 1.0

        assert not self.open_paths
        log(f"loop_count: {self.loop_count}")
        log(f"arc_fail_count: {self.arc_fail_count}")
        log(f"len(path): {len(self.path)}")
        log(f"path_fail_count: {self.path_fail_count}")

        self.done_generating()

    def _filter_arc(self, arc: ArcData) -> Optional[ArcData]:
        """
        Remove any arc that is very close to the edge of the part in it's entirety.
        """
        if len(arc.path.coords) < 3:
            return None

        if arc.path.length <= self.step / 20:
            # Arc too short to care about.
            return None

        if not arc.path.intersects(self.polygon):
            return None

        poly_arc = Polygon(arc.path)
        for ring in self.dilated_polygon_boundaries:
            if ring.contains(poly_arc):
                return None

        return arc

    def _start_point_perimeter(
            self, polygons: MultiPolygon, already_cut: Polygon, step: float) -> VoronoiCenters:
        """
        Recalculate the start point to be inside the cut area, adjacent to the perimeter.
        """
        starting_radius = self.starting_radius or step
        voronoi = VoronoiCenters(polygons, preserve_edge=True)

        perimiter_point = voronoi.start_point
        assert perimiter_point is not None
        voronoi_edge_index = voronoi.vertex_to_edges[perimiter_point.coords[0]]
        assert len(voronoi_edge_index) == 1
        voronoi_edge = voronoi.edges[voronoi_edge_index[0]]
        cut_edge_section = already_cut.intersection(voronoi_edge)
        if cut_edge_section.length > starting_radius * 2:
            new_start_point = cut_edge_section.interpolate(0.5, normalized=True)

            voronoi = VoronoiCenters(polygons, starting_point=new_start_point)
        else:
            # Can't fit starting_radius here.
            # Look for widest point in already_cut area.
            cut_voronoi = VoronoiCenters(already_cut, preserve_widest=True)
            voronoi = VoronoiCenters(polygons, starting_point=cut_voronoi.start_point)
            del cut_voronoi

        return voronoi


class EntryCircle(BaseGeometry):
    center: Point
    radius: float
    start_angle: float
    path: List[Union[ArcData, LineData]]

    def __init__(
            self,
            to_cut: Polygon,
            center: Point,
            radius: float,
            step: float,
            winding_dir: ArcDir,
            start_angle: float = 0,
            already_cut: Optional[Polygon] = None,
            path: Optional[List[Union[ArcData, LineData]]] = None) :
        super().__init__(to_cut, step, winding_dir, already_cut)

        self.center = center
        self.radius = radius
        self.start_angle = start_angle

        if path is None:
            path = []
        self.path = path

    def spiral(self):
        loop: float = 0.25
        offset: List[float] = [0.0, 0.0]
        new_arcs = []
        mask = self.center.buffer(self.radius)
        not_done = True
        while loop * self.step <= self.radius and not_done:
            orientation = round(loop * 4) % 4
            if orientation == 0:
                start_angle = 0
                offset[1] -= self.step / 4
                section_radius = loop * self.step
            elif orientation == 1:
                start_angle = math.pi / 2
                offset[0] -= self.step / 4
                section_radius = loop * self.step
            elif orientation == 2:
                start_angle = math.pi
                offset[1] += self.step / 4
                section_radius = loop * self.step
            elif orientation == 3:
                start_angle = 3 * math.pi / 2
                offset[0] += self.step / 4
                section_radius = loop * self.step
            else:
                raise

            section_center = Point(self.center.x + offset[0], self.center.y + offset[1])
            new_arc = create_arc(
                    section_center, section_radius, start_angle, math.pi / 2, ArcDir.CW)

            if new_arc.path.intersects(mask.exterior):
                # Spiral has crossed bounding circle.
                masked_arc = new_arc.path.intersection(mask)
                if masked_arc.geom_type == "MultiLineString":
                    longest = None
                    for arc in list(masked_arc.geoms):
                        if not longest or arc.length > longest.length:
                            longest = arc
                    masked_arc = longest
                if masked_arc.geom_type == "GeometryCollection":
                    longest = None
                    for arc in list(masked_arc.geoms):
                        if not longest or arc.length > longest.length:
                            longest = arc
                    masked_arc = longest

                new_arc = create_arc_from_path(
                        new_arc.origin, masked_arc, new_arc.radius, ArcDir.CW)
                new_arc = complete_arc(new_arc, new_arc.winding_dir)

                not_done = False

            if self.winding_dir == ArcDir.CCW:
                new_arc = mirror_arc(self.center.x, new_arc)

            new_arcs.append(new_arc)
            loop += 0.25  # 1/4 turn.

        #self._queue_arcs(new_arcs)

        sorted_arcs = self._split_arcs(new_arcs)
        sorted_arcs = [complete_arc(sorted_arc) for sorted_arc in sorted_arcs]
        self._queue_arcs(sorted_arcs)

        self._flush_arc_queues()

    def circle(self):
        if self.winding_dir == ArcDir.CCW:
            angle = -math.pi * 1.99999
        else:
            angle = math.pi * 1.99999

        new_arc = create_arc(self.center, self.radius, 0, angle, self.winding_dir)
        sorted_arcs = self._split_arcs([new_arc])
        sorted_arcs = [complete_arc(sorted_arc) for sorted_arc in sorted_arcs]
        self._queue_arcs(sorted_arcs)
        self._flush_arc_queues()

